import shlex
from typing import Dict, Optional, Any 
from stimela.kitchen.cab import Cab
from stimela.kitchen.batch import Batch
from stimela import logger
from stimela.utils.xrun_poll import xrun
from stimela.exceptions import StimelaCabRuntimeError
import subprocess


binary = "srun"

class SlurmBatch(Batch):

    def exists(self):
        try:
            subprocess.check_output(["which", binary])
        except subprocess.CalledProcessError:
            raise SystemError(f"The '{binary}' tool could not be found on the system. Is 'slurm' installed?")

    # This funtion can only be called after the __init_cab__()
    def submit(self, jobfile, runcmd, *args):
        self.exists()
        jobname = f"{self.name}.job"
        with open(jobfile, "w") as fh:
            fh.writelines("#!/bin/bash\n")
            fh.writelines(f"#SBATCH --job-name={jobname}\n")
            fh.writelines(f"#SBATCH --qos=normal\n")
            fh.writelines(f"#SBATCH --mail-type=ALL\n")
            fh.writelines(f"#SBATCH --error=stimela_slurm_{self.name}.err\n")
            #fh.writelines(f"#SBATCH --time=2-00:00\n")
            if self.mem:
                fh.writelines(f"#SBATCH --mem={self.mem}\n")
            if self.cpus:
                fh.writelines(f"#SBATCH --cpus-per-task={self.cpus}\n")
            if self.email:
                fh.writelines(f"#SBATCH --mail-user={self.email}\n")
            for arg in args:
                fh.writelines(f"{arg}\n")

            fh.writelines("\n")
            fh.writelines(f"{runcmd}\n")

        self.log.info(f"Submiting job to schedular. The job name is {jobname}")
        retcode = xrun(binary, [jobfile], shell=False, log=self.log, 
                    output_wrangler=self.cab.apply_output_wranglers, 
                    return_errcode=True, command_name=binary)

        # if retcode is not 0, and cab didn't declare itself a success,
        if retcode:
            if not self.cab.runtime_success:
                raise StimelaCabRuntimeError(f"{binary} returned non-zero exit status {retcode}", log=self.log)
        else:
            if self.cab.runtime_success is False:
                raise StimelaCabRuntimeError(f"{binary} was marked as failed based on its output", log=self.log)

        return retcode



from dataclasses import dataclass
import logging, time, json, datetime, yaml, os.path, uuid, pathlib

from typing import Dict, List, Optional, Any
from enum import Enum
from collections import OrderedDict

from omegaconf import OmegaConf, DictConfig, ListConfig
from scabha.basetypes import EmptyDictDefault, EmptyListDefault

import stimela
from stimela.kitchen.cab import Cab
from stimela.kitchen.step import Step
from stimela.utils.xrun_asyncio import xrun, dispatch_to_log
from stimela.exceptions import StimelaCabParameterError, StimelaCabRuntimeError, CabValidationError
from stimela.stimelogging import log_exception
from stimela.backends import resolve_required_mounts
# these are used to drive the status bar
from stimela.stimelogging import declare_subcommand, declare_subtask, update_process_status

import kubernetes
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream
from dask_kubernetes import make_pod_spec, KubeCluster

@dataclass
class KubernetesDaskRuntime(object):
    num_workers: int = 0
    cpu_limit: int = 1
    memory_limit: Optional[str] = None
    threads_per_worker: int = 1
    name: Optional[str] = None
    persist: bool = False

# dict of methods for converting an object to text format
_InjectedFileFormatters = dict(
    yaml = yaml.dump,
    json = json.dumps,
    txt = str
)

InjectedFileFormats = Enum("InjectedFileFormats", " ".join(_InjectedFileFormatters.keys()), module=__name__)

@dataclass
class KubernetesFileInjection(object):
    format: InjectedFileFormats = "txt"
    content: Any = ""

@dataclass
class KubernetesLocalMount(object):
    path: str
    dest: str = ""              # destination path -- same as local if empty
    readonly: bool = False      # mount as readonly, but it doesn't work (yet?)
    mkdir: bool = False         # create dir, if it is missing

@dataclass
class KubernetesRuntime(object):
    namespace: str
    dask_cluster: KubernetesDaskRuntime = KubernetesDaskRuntime()
    inject_files: Dict[str, KubernetesFileInjection] = EmptyDictDefault()
    pre_commands: List[str] = EmptyListDefault()
    local_mounts: Dict[str, KubernetesLocalMount] = EmptyDictDefault()
    env: Dict[str, str] = EmptyDictDefault()
    run_dir: str = "."          # directory to run in inside container

KubernetesRuntimeSchema = OmegaConf.structured(KubernetesRuntime)

_kube_client = _kube_config = None

def get_kube_api():
    global _kube_client
    global _kube_config

    if _kube_config is None:
        _kube_config = True
        kubernetes.config.load_kube_config()

    return core_v1_api.CoreV1Api()



def run(cab: Cab, params: Dict[str, Any], runtime: Dict[str, Any], fqname: str,
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None, batch=None):
    """Runs cab contents

    Args:
        cab (Cab): cab object
        log (logger): logger to use
        subst (Optional[Dict[str, Any]]): Substitution dict for commands etc., if any.

    Returns:
        Any: return value (e.g. exit code) of content
    """
    if not cab.image:
        raise StimelaCabRuntimeError(f"kubernetes runner requires cab.image to be set")

    if 'kube' in stimela.CONFIG.opts.runtime:
        kube = OmegaConf.merge(KubernetesRuntimeSchema, stimela.CONFIG.opts.runtime.kube)

    # get kube info from cab's runtime section
    if 'kube' in cab.runtime:
        kube = OmegaConf.merge(KubernetesRuntimeSchema, cab.runtime.kube)

    # augment with step-specific one
    if 'kube' in runtime:
        kube = OmegaConf.merge(KubernetesRuntimeSchema, runtime.kube)

    namespace = kube.namespace
    if not namespace:
        raise StimelaCabRuntimeError(f"runtime.kube.namespace must be set")

    args = cab.flavour.get_arguments(cab, params, subst)
    log.debug(f"command line is {args}")

    cabstat = cab.reset_status()

    command_name = cab.flavour.command_name

    cluster = podname = cluster_name = None
    kube_api = get_kube_api()

    start_time = datetime.datetime.now()
    def elapsed(since=None):
        """Returns string representing elapsed time"""
        return str(datetime.datetime.now() - (since or start_time)).split('.', 1)[0]

    last_update = 0
    def update_status():
        nonlocal last_update
        if time.time() - last_update >= 1:
            update_process_status()
            last_update = time.time()

    numba_cache_dir = os.path.expanduser("~/.cache/numba")
    pathlib.Path(numba_cache_dir).mkdir(parents=True, exist_ok=True)

    with declare_subtask(f"kubernetes:{os.path.basename(command_name)}"):
        try:
            if kube.dask_cluster.num_workers:
                cluster_name = kube.dask_cluster.name
                if kube.dask_cluster.persist:
                    try:
                        resp = kube_api.read_namespaced_service(name=cluster_name, namespace=namespace)
                        cluster = True
                        log.info(f"persistent dask cluster {cluster_name} appears to be up")
                    except ApiException as exc:
                        log.info(f"persistent dask cluster {cluster_name} is not up")

                if cluster is None:
                    with declare_subcommand("starting dask cluster"):
                        log.info(f"starting dask cluster {cluster_name} for {command_name}")
                        pod_spec = make_pod_spec(image=cab.image,
                                                cpu_limit=kube.dask_cluster.cpu_limit,
                                                memory_limit=kube.dask_cluster.memory_limit,
                                                threads_per_worker=kube.dask_cluster.num_workers)

                        cluster = KubeCluster(pod_spec, name=cluster_name, namespace=namespace, shutdown_on_close=not kube.dask_cluster.persist)
                        update_status()
                        cluster.scale(kube.dask_cluster.num_workers)
                        update_status()

            podname = fqname.replace(".", "--").replace("_", "--") + "--" + uuid.uuid4().hex

            pod_manifest = dict(
                apiVersion  =  'v1',
                kind        =  'Pod',
                metadata    = dict(name = podname),
                spec        = dict(
                    containers = [dict(
                            image   = cab.image,
                            name    = podname,
                            args    = ["/bin/sh", "-c", "while true;do date;sleep 5; done"],
                            env     = [],
                            securityContext = dict(
                                    runAsNonRoot = True,
                                    runAsUser = os.getuid(),
                                    runAsGroup = os.getgid()
                            ),
                            volumeMounts = []
                    )],
                    volumes = []
                )
            )
            # add runtime env settings
            for name, value in kube.env.items():
                value = os.path.expanduser(value)
                pod_manifest['spec']['containers'][0]['env'].append(dict(name=name, value=value))

            # add local mounts
            def add_local_mount(name, path, dest, readonly):
                name = name.replace("_", "-")  # sanizitze name
                pod_manifest['spec']['volumes'].append(dict(
                    name = name,
                    hostPath = dict(path=path, type='Directory' if os.path.isdir(path) else 'File')
                ))
                pod_manifest['spec']['containers'][0]['volumeMounts'].append(dict(name=name, mountPath=dest, readOnly=readonly))

            # this will accumulate mounted paths from runtime spec
            prior_mounts = {}

            # add local mounts from runtime spec
            for name, mount in kube.local_mounts.items():
                path = os.path.abspath(os.path.expanduser(mount.path))
                dest = os.path.abspath(os.path.expanduser(mount.dest)) if mount.dest else path
                if not os.path.exists(path) and mount.mkdir:
                    pathlib.Path(path).mkdir(parents=True)
                add_local_mount(name, path, dest, mount.readonly)
                if path == dest:
                    prior_mounts[path] = not mount.readonly

            # add local mounts to support parameters
            req_mounts = resolve_required_mounts(params, cab.inputs, cab.outputs, prior_mounts=prior_mounts)
            for i, (path, readwrite) in enumerate(req_mounts.items()):
                log.info(f"adding local mount {path} (readwrite={readwrite})")
                add_local_mount(f"automount-{i}", path, path, not readwrite)

            # start pod and wait for it to come up
            with declare_subcommand("starting pod"):
                log.info(f"starting pod {podname} for {command_name}")
                resp = kube_api.create_namespaced_pod(body=pod_manifest, namespace=namespace)
                log.debug(f"create_namespaced_pod({podname}): {resp}")

                while True:
                    update_status()
                    resp = kube_api.read_namespaced_pod(name=podname, namespace=namespace)
                    log.debug(f"read_namespaced_pod({podname}): {resp}")
                    if resp.status.phase != 'Pending':
                        break
                    time.sleep(.5)
                log.info(f"  pod started after {elapsed()}")


            def run_pod_command(command, cmdname, input=None, wrangler=None):
                if type(command) is str:
                    command = ["/bin/sh", "-c", command]
                has_input = bool(input)

                resp = stream(kube_api.connect_get_namespaced_pod_exec, podname, namespace,
                            command=command,
                            stderr=True, stdin=has_input,
                            stdout=True, tty=False,
                            _preload_content=False)

                while resp.is_open():
                    update_status()
                    resp.update(timeout=1)
                    if resp.peek_stdout():
                        for line in resp.read_stdout().rstrip().split("\n"):
                            dispatch_to_log(log, line, cmdname, "stdout",
                                            output_wrangler=wrangler)
                    if resp.peek_stderr():
                        for line in resp.read_stderr().rstrip().split("\n"):
                            dispatch_to_log(log, line, cmdname, "stderr",
                                            output_wrangler=wrangler)
                    if has_input:
                        if input:
                            resp.write_stdin(input)
                            input = None
                        else:
                            break

                retcode = resp.returncode
                resp.close()
                return retcode

            # inject files into pod
            if 'inject_files' in kube:
                with declare_subcommand("configuring pod (inject)"):
                    for filename, injection in kube.inject_files.items_ex():
                        content = injection.content
                        formatter = _InjectedFileFormatters.get(injection.format.name)
                        if formatter is None:
                            raise StimelaCabParameterError(f"unsupported format {injection.format.name} for {filename}")
                        # convert content to something serializable
                        if isinstance(content, (DictConfig, ListConfig)):
                            content = OmegaConf.to_container(content)
                        content = formatter(content)
                        log.info(f"injecting {filename} into pod {podname}")
                        retcode = run_pod_command(f"mkdir -p {os.path.dirname(filename)}; cat >{filename}", "inject", input=content)
                        if retcode:
                            log.warning(f"injection returns exit code {retcode} after {elapsed()}")
                        else:
                            log.info(f"injection successful after {elapsed()}")


            if 'pre_commands' in kube:
                with declare_subcommand("configuring pod (pre-commands)"):
                    for pre_command in kube.pre_commands:
                        log.info(f"running pre-command '{pre_command}' in pod {podname}")
                        # calling exec and waiting for response
                        retcode = run_pod_command(pre_command, pre_command.split()[0])
                        if retcode:
                            log.warning(f"pre-command returns exit code {retcode} after {elapsed()}")
                        else:
                            log.info(f"pre-command successful after {elapsed()}")

            # do we need to chdir
            if kube.run_dir:
                rundir = os.path.abspath(kube.run_dir)
                args = ["python", "-c", f"import os,sys; os.chdir('{rundir}'); os.execlp('{args[0]}', *sys.argv[1:])"] + list(args)

            log.info(f"running {command_name} in pod {podname}")
            with declare_subcommand(os.path.basename(command_name)):
                retcode = run_pod_command(args, command_name, wrangler=cabstat.apply_wranglers)

            # check if output marked it as a fail
            if cabstat.success is False:
                log.error(f"declaring '{command_name}' as failed based on its output")

            # if retcode != 0 and not explicitly marked as success, mark as failed
            if retcode and cabstat.success is not True:
                cabstat.declare_failure(f"{command_name} returns error code {retcode} after {elapsed()}")
            else:
                log.info(f"{command_name} returns exit code {retcode} after {elapsed()}")

            return cabstat

        # handle various failure modes by logging errors appropriately
        except KeyboardInterrupt:
            log.error(f"kubernetes invocation of {command_name} interrupted with Ctrl+C after {elapsed()}")
            raise StimelaCabRuntimeError(f"{command_name} interrupted with Ctrl+C after {elapsed()}")
        except ApiException as exc:
            if exc.body:
                exc = (exc, json.loads(exc.body))
            log.error(f"kubernetes invocation of {command_name} failed with an ApiException after {elapsed()}")
            raise StimelaCabRuntimeError("kubernetes API error", exc)
        # this drops out as a normal error response
        except StimelaCabRuntimeError as exc:
            log.error(f"kubernetes invocation of {command_name} failed after {elapsed()}")
            raise
        except Exception as exc:
            log.error(f"kubernetes invocation of {command_name} failed after {elapsed()}")
            raise StimelaCabRuntimeError("kubernetes backend error", exc)

        # cleanup
        finally:
            if cluster and not kube.dask_cluster.persist:
                update_status()
                log.info(f"stopping dask cluster {cluster_name}")
                log.info(f"cluster logs: {cluster.get_logs()}")
                cluster.close()
            if podname:
                try:
                    update_status()
                    log.info(f"deleting pod {podname}")
                    resp = kube_api.delete_namespaced_pod(name=podname, namespace=namespace)
                    log.debug(f"delete_namespaced_pod({podname}): {resp}")
                    # while True:
                    #     resp = kube_api.read_namespaced_pod(name=podname, namespace=namespace)
                    #     log.debug(f"read_namespaced_pod({podname}): {resp}")
                    #     log.info(f"  pod phase is {resp.status.phase} after {elapsed()}")
                    #     time.sleep(.5)
                except ApiException as exc:
                    body = json.loads(exc.body)
                    log_exception(StimelaCabRuntimeError(f"kubernetes API error while deleting pod {podname}", (exc, body)), severity="warning")
                except Exception as exc:
                    log_exception(StimelaCabRuntimeError(f"kubernetes API error while deleting pod {podname}: {exc}"), severity="warning")


# kubectl -n rarg get pods -A
# kubectl -n rarg delete service recipetestqcdaskcluster
# kubectl -n rarg delete poddisruptionbudget recipetestqcdaskcluster
# kubectl -n rarg port-forward service/qc-test-cluster 18787:http-dashboard
# kubectl -n rarg logs pod_id
"""
https://kubernetes.dask.org/en/latest/kubecluster.html#dask_kubernetes.KubeCluster

We recommend adding the --death-timeout, '60' arguments and the restartPolicy: Never attribute
to your worker specification. This ensures that these pods will clean themselves up if your Python process disappears unexpectedly.

OMS: this should be set in the structure returned by make_pod_spec(). Seems to be set by default.

https://kubernetes.dask.org/en/latest/testing.ht

By default we set the --keep-cluster flag in setup.cfg which means the Kubernetes container will persist between pytest runs to
avoid creation/teardown time. Therefore you may want to manually remove the container when you are done working on dask-kubernetes:
"""



import os, re, subprocess, pytest
from .test_recipe import change_test_dir, run, verify_output

def callable_function(a: int, b: str):
    print(f"callable_function({a},'{b}')")
    return a*2

def callable_function_dict(a: int, b: str):
    print(f"callable_function_dict({a},'{b}')")
    return dict(x=a*2, y=b+b)

def test_wrangler_replace_suppress():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -v run test_callables.yml")
    assert retcode == 0
    print(output)
    assert verify_output(output, 'y = 46barbar')



import os.path
import uuid
from dataclasses import make_dataclass
from typing import Any, List, Dict, Optional, OrderedDict, Union, Callable
from omegaconf.omegaconf import OmegaConf, DictConfig, ListConfig
from omegaconf.errors import OmegaConfBaseException
from yaml.error import YAMLError

from scabha.exceptions import ScabhaBaseException
from .deps import ConfigDependencies
from .resolvers import resolve_config_refs
from .cache import load_cache, save_cache
from .common import *


def load(path: str, use_sources: Optional[List[DictConfig]] = [], name: Optional[str]=None, 
        location: Optional[str]=None, 
        includes: bool=True, selfrefs: bool=True, include_path: str=None, 
        use_cache: bool = True,
        no_toplevel_cache = False,
        verbose: bool = False):
    """Loads config file, using a previously loaded config to resolve _use references.

    Args:
        path (str): path to config file
        use_sources (Optional[List[DictConfig]]): list of existing configs to be used to resolve "_use" references, 
                or None to disable
        name (Optional[str]): name of this config file, used for error messages
        location (Optional[str]): location where this config is being loaded (if not at root level)
        includes (bool, optional): If True (default), "_include" references will be processed
        selfrefs (bool, optional): If False, "_use" references will only be looked up in existing config.
            If True (default), they'll also be looked up within the loaded config.
        include_path (str, optional):
            if set, path to each config file will be included in the section as element 'include_path'

    Returns:
        Tuple of (conf, dependencies)
            conf (DictConfig): config object    
            dependencies (ConfigDependencies): filenames that were _included
    """
    use_toplevel_cache = use_cache and not no_toplevel_cache
    conf, dependencies = load_cache((path,), verbose=verbose) if use_toplevel_cache else (None, None)

    if conf is None:
        subconf = OmegaConf.load(path)
        name = name or os.path.basename(path)
        dependencies = ConfigDependencies()
        dependencies.add(path)
        # include ourself into sources, if _use is in effect, and we've enabled selfrefs
        if use_sources is not None and selfrefs:
            use_sources = [subconf] + list(use_sources)
        conf, deps = resolve_config_refs(subconf, pathname=path, location=location, name=name, 
                            includes=includes, use_cache=use_cache, use_sources=use_sources, include_path=include_path)
        # update overall dependencies
        dependencies.update(deps)

        # # check for missing requirements
        # dependencies.scan_requirements(conf, location, path)

        if use_cache:
            save_cache((path,), conf, dependencies, verbose=verbose)

    return conf, dependencies

def load_nested(filelist: List[str], 
                structured: Optional[DictConfig] = None,
                typeinfo = None,
                use_sources: Optional[List[DictConfig]] = [],
                location: Optional[str] = None,  
                nameattr: Union[Callable, str, None] = None,
                config_class: Optional[str] = None,
                include_path: Optional[str] = None,
                use_cache: bool = True,
                verbose: bool = False):
    """Builds nested configuration from a set of YAML files corresponding to sub-sections

    Parameters
    ----------
    conf : OmegaConf object
        root OmegaConf object to merge content into
    filelist : List[str]
        list of subsection config files to load
    schema : Optional[DictConfig]
        schema to be applied to each file, if any
    use_sources : Optional[List[DictConfig]]
        list of existing configs to be used to resolve "_use" references, or None to disable
    location : Optional[str]
        if set, contents of files are being loaded under 'location.subsection_name'. If not set, then 'subsection_name' is being
        loaded at root level. This is used for correctly formatting error messages and such.
    nameattr : Union[Callable, str, None]
        if None, subsection_name will be taken from the basename of the file. If set to a string such as 'name', will set 
        subsection_name from that field in the subsection config. If callable, will be called with the subsection config object as a single 
        argument, and must return the subsection name
    config_class : Optional[str]
        name of config dataclass to form (when using typeinfo), if None, then generated automatically
    include_path : Optional[str] 
        if set, path to each config file will be included in the section as element 'include_path'

    Returns
    -------
        Tuple of (conf, dependencies)
            conf (DictConfig): config object    
            dependencies (set): set of filenames that were _included

    Raises
    ------
    NameError
        If subsection name is not resolved
    """
    section_content, dependencies = load_cache(filelist, verbose=verbose) if use_cache else (None, None)

    if section_content is None:
        section_content = {} # OmegaConf.create()
        dependencies = ConfigDependencies()

        for path in filelist:
            # load file
            subconf, deps = load(path, location=location, use_sources=use_sources, include_path=include_path)
            dependencies.update(deps)
            if include_path:
                subconf[include_path] = path

            # figure out section name
            if nameattr is None:
                name = os.path.splitext(os.path.basename(path))[0]
            elif callable(nameattr):
                name = nameattr(subconf) 
            elif nameattr in subconf:
                name = subconf.get(nameattr)
            else:
                raise NameError(f"{path} does not contain a '{nameattr}' field")

            # # resolve _use and _include statements
            # try:
            #     subconf = resolve_config_refs(subconf, f"{location}.{name}" if location else name, conf, subconf))
            # except (OmegaConfBaseException, YAMLError) as exc:
            #     raise ConfigurattError(f"config error in {path}: {exc}")

            # apply schema
            if structured is not None:
                try:
                    subconf = OmegaConf.merge(structured, subconf) 
                except (OmegaConfBaseException, YAMLError) as exc:
                    raise ConfigurattError(f"schema error in {path}: {exc}")

            section_content[name] = subconf

        if structured is None and typeinfo is not None:
            if config_class is None:
                config_class = "ConfigClass_" + uuid.uuid4().hex
            fields = [(name, typeinfo) for name in section_content.keys()]
            datacls = make_dataclass(config_class, fields)
            # datacls.__module__ == __name__  # for pickling
            structured = OmegaConf.structured(datacls)
            section_content = OmegaConf.merge(structured, section_content)
    
        if use_cache:
            save_cache(filelist, section_content, dependencies, verbose=verbose)

    return section_content, dependencies


def check_requirements(conf: DictConfig, bases: List[DictConfig], strict: bool = True):
    # build requirements map first using recursive helper
    def _check_item(section, name):
        """Helper function, checks if a dotted name x.y.z is in conf, or .x is in section"""
        name_elems = name.split(".")
        # if name starts with ".", look in current section, else check config and any bases
        if name_elems[0] == "":
            name_elems = name_elems[1:]
            configs = [section]
        else:
            configs = [conf] + bases
        # try all configs
        for config in configs:
            # start looking at this config's top level
            section = config
            for elem in name_elems:
                # break loop if not found at this level
                if elem not in section:
                    break
                try:
                    section = section[elem]
                except Exception as exc:
                    break
            # got to the end? Success
            else:
                return True
        # nothing found -- return False
        return False

    def _scan(section, location=""):
        # Short-circuit out if section is empty. Seems both DictConfig and ListConfig can be a kind of None
        # (funny OmegeConf API feature), where they're an instance of DictConfig or ListConfig, but don't support __getitem__. 
        #
        if not section:
            return [], False

        delete_self = False
        to_delete = []
        unresolved = []
        if isinstance(section, DictConfig):
            # get requirements from section
            reqs = pop_conf(section, "_requires", [])
            contingents = pop_conf(section, "_contingent", [])
            if type(reqs) is str:
                reqs = [reqs]
            if type(contingents) is str:
                contingents = [contingents]
            if not isinstance(reqs, (list, ListConfig)):
                raise ConfigurattError(f"'{location or ''}._requires' must be a string or a sequence")
            if not isinstance(contingents, (list, ListConfig)):
                raise ConfigurattError(f"'{location or ''}._contingent' must be a string or a sequence")
            # check requirements and contingencies
            for req in reqs:
                if not _check_item(section, req):
                    unresolved.append((location, req, ConfigurattError(f"section '{location}' has missing requirement '{req}'")))
            for cont in contingents:
                if not _check_item(section, cont):
                    unresolved.append((location, cont, None))
                    delete_self = True
            # recurse into content
            for name, value in section.items_ex(resolve=False):
                unres, delete = _scan(value, location=f"{location}.{name}" if location else name)
                unresolved += unres
                if delete:
                    to_delete.append(name)
            # delete what needs to be deleted
            for name in to_delete:
                del section[name]
        # lists -- recurse into content
        elif isinstance(section, ListConfig):
            for i, value in enumerate(section._iter_ex(resolve=False)):
                unres, delete = _scan(value, location=f"{location or ''}[{i}]")
                unresolved += unres
                if delete:
                    to_delete.append(i)
            # delete what needs to be deleted
            for i in to_delete[::-1]:
                del section[i]
            delete_self = False
        return unresolved, delete_self
    # build list of unresolved items
    unresolved, delete = _scan(conf)

    # top-level config wants to be deleted? this is an error, it's not supposed to be contingent
    if delete:
        raise ConfigurattError("top-level configuration is contingent on missing sections")

    # in strict mode, throw any unresolved errors
    if strict:
        errors = [exc for (_, _, exc) in unresolved if exc is not None]
        if errors:
            raise ConfigurattError("configuration has missing requirements", nested=errors)

    return unresolved



from scabha.exceptions import *

class StimelaBaseException(ScabhaBaseException):
    pass

class BackendError(StimelaBaseException):
    pass

class CabValidationError(StimelaBaseException):
    pass

class SchemaError(StimelaBaseException):
    pass

class DefinitionError(StimelaBaseException):
    pass

class StepValidationError(StimelaBaseException):
    pass

class StepSelectionError(StimelaBaseException):
    pass

class RecipeValidationError(StimelaBaseException):
    pass

class CabValidationError(StimelaBaseException):
    pass

class StimelaCabParameterError(StimelaBaseException):
    pass

class StimelaCabError(StimelaBaseException):
    pass

class StimelaStepExecutionError(StimelaBaseException):
    pass

class StimelaRecipeExecutionError(StimelaBaseException):
    pass

class StimelaBaseImageError(StimelaBaseException):
    pass

class StimelaRuntimeError(StimelaBaseException):
    pass

class StimelaCabRuntimeError(StimelaBaseException):
    pass

class StimelaCabOutputError(StimelaBaseException):
    pass

class StimelaProcessRuntimeError(StimelaBaseException):
    pass


class PipelineException(StimelaBaseException):
    """ 
    Encapsulates information about state of pipeline when an
    exception occurs
    """

    def __init__(self, exception, completed, failed, remaining):
        message = ("Job '%s' failed: %s" % (failed.label, str(exception)))

        super(PipelineException, self).__init__(message)

        self._completed = completed
        self._failed = failed
        self._remaining = remaining

    @property
    def completed(self):
        return self._completed

    @property
    def failed(self):
        return self._failed

    @property
    def remaining(self):
        return self._remaining




import os, os.path, re, fnmatch, copy, traceback
from typing import Any, Tuple, List, Dict, Optional, Union
from dataclasses import dataclass
from omegaconf import MISSING, OmegaConf, DictConfig, ListConfig
from collections import OrderedDict
from collections.abc import Mapping
import rich.table

from concurrent.futures import ProcessPoolExecutor

from stimela.config import EmptyDictDefault, EmptyListDefault
import stimela
from stimela import log_exception, stimelogging
from stimela.exceptions import *
from scabha.basetypes import SkippedOutput
from scabha.validate import evaluate_and_substitute, Unresolved, join_quote
from scabha.substitutions import SubstitutionNS
from scabha.cargo import Parameter, Cargo, ParameterCategory
from scabha.basetypes import File, Directory, MS, UNSET, Placeholder
from .cab import Cab
from .batch import Batch
from .step import Step
from stimela import task_stats 


class DeferredAlias(Unresolved):
    """Class used as placeholder for deferred alias lookup (i.e. before an aliased value is available)"""
    pass


@dataclass
class ForLoopClause(object):
    # name of list variable
    var: str 
    # This should be the name of an input that provides a list, or a list
    over: Optional[Any] = None
    # If !=0 , this is a scatter not a loop -- things may be evaluated in parallel using this many workers
    # (use -1 to scatter to unlimited number of workers)
    scatter: int = 0
    # How to indicate the status of the loop on the console.
    # Default is "i/N", where i is the current index plus 1, and N is the total number of loops. 
    # A format string can be supplied instead.
    display_status: Optional[str] = None

def IterantPlaceholder(name: str):
    return name

@dataclass
class Recipe(Cargo):
    """Represents a sequence of steps.

    Additional attributes available after validation with arguments are as per for a Cab:

        self.input_output:      combined parameter dict (self.input + self.output), maps name to Parameter

    Raises:
        various classes of validation errors
    """
    steps: Dict[str, Step] = EmptyDictDefault()     # sequence of named steps

    assign: Dict[str, Any] = EmptyDictDefault()     # assigns variables

    assign_based_on: Dict[str, Any] = EmptyDictDefault()
                                                    # assigns variables based on values of other variables

    aliases: Dict[str, Any] = EmptyDictDefault()

    # make recipe a for_loop-gather (i.e. parallel for loop)
    for_loop: Optional[ForLoopClause] = None

    # logging control, overrides opts.log.init_logname and opts.log.logname 
    init_logname: Optional[str] = None
    logname: Optional[str] = None
    batch: Optional[Batch] = None
    
    # # if not None, do a while loop with the conditional
    # _while: Conditional = None
    # # if not None, do an until loop with the conditional
    # _until: Conditional = None

    def __post_init__ (self):
        Cargo.__post_init__(self)
        # flatten aliases and assignments
        self.aliases = self.flatten_param_dict(OrderedDict(), self.aliases)
        # check that schemas are valid
        for io in self.inputs, self.outputs:
            for name, schema in io.items():
                if not schema:
                    raise RecipeValidationError(f"recipe '{self.name}': '{name}' does not define a valid schema")
        # check for repeated aliases
        for name, alias_list in self.aliases.items():
            if name in self.inputs_outputs:
                raise RecipeValidationError(f"recipe '{self.name}': alias '{name}' also appears under inputs or outputs")
            if type(alias_list) is str:
                alias_list = self.aliases[name] = [alias_list]
            if not hasattr(alias_list, '__iter__') or not all(type(x) is str for x in alias_list):
                raise RecipeValidationError(f"recipe '{self.name}': alias '{name}': name or list of names expected")
            for x in alias_list:
                if '.' not in x:
                    raise RecipeValidationError(f"recipe '{self.name}': alias '{name}': invalid target '{x}' (missing dot)")
        # instantiate steps if needed (when creating from an omegaconf)
        if type(self.steps) is not OrderedDict:
            steps = OrderedDict()
            for label, stepconfig in self.steps.items():
                stepconfig.name = label
                stepconfig.fqname = f"{self.name}.{label}"
                try:
                    step = OmegaConf.unsafe_merge(StepSchema.copy(), stepconfig)
                    steps[label] = Step(**step)
                except Exception as exc:
                    raise StepValidationError(f"recipe '{self.name}': error in definition of step '{label}'", exc)
            self.steps = steps
        # check that assignments don't clash with i/o parameters
 
        self.validate_assignments(self.assign, self.assign_based_on, self.name)

        # check that for-loop variable does not clash
        if self.for_loop:
            for io, io_label in [(self.inputs, "input"), (self.outputs, "output")]:
                if self.for_loop.var in io:
                    raise RecipeValidationError(f"recipe '{self.name}': for_loop.var={self.for_loop.var} clashes with an {io_label} parameter")
        # marked when finalized
        self._alias_map  = None
        # set of keys protected from assignment
        self._protected_from_assign = set()
        self._for_loop_values = self._for_loop_scatter = None
        # process pool used to run for-loops
        self._loop_pool = None

    def validate_assignments(self, assign, assign_based_on, location):
        # collect a list of all assignments
        assignments = OrderedDict()
        for key in assign:
            assignments[key] = "assign"
        for basevar, lookup_list in assign_based_on.items():
            if not isinstance(lookup_list, Mapping):
                raise RecipeValidationError(f"{location}.{assign_based_on}.{basevar}: mapping expected")
            # for assign_list in lookup_list.values():
            #     for key in assign_list:
            #         assignments[key] = f"assign_based_on.{basevar}"
        # # check that none clash
        # for key, assign_label in assignments.items():
        #     for io, io_label in [(self.inputs, "input"), (self.outputs, "output")]:
        #         if key in io:
        #             raise RecipeValidationError(f"'{location}.{assign_label}.{key}' clashes with an {io_label}")

    def update_assignments(self, subst: SubstitutionNS, whose = None, params: Dict[str, Any] = {}, 
                            ignore_subst_errors: bool = False):
        """Updates variable assignments, using the recipe's (or a step's) 'assign' and 'assign_based_on' sections.
        Also updates the corresponding (recipe or step's) file logger.

        Args:
            subst (SubstitutionNS): substitution namespace
            whose (Step or None): if None, use recipe's (self) assignments, else use this step's
            params (dict, optional): dictionary of parameters 
            ignore_subst_errors (bool): ignore substitution errors (default is False)

        Raises:
            AssignmentError: on errors
        """
        whose = whose or self
        # short-circuit out if nothing to do
        if not whose.assign and not whose.assign_based_on:
            return

        def flatten_dict(input_dict, output_dict={},  prefix=""):
            for name, value in input_dict.items():
                name = f"{prefix}{name}"
                if isinstance(value, (dict, OrderedDict, DictConfig)):
                    flatten_dict(value, output_dict=output_dict, prefix=f"{name}.")
                else:
                    output_dict[name] = value
            return output_dict

        # accumulate all assignments for a final round of evaluation
        assign = {}

        def do_assign(assignments):
            """Helper function to process a list of assignments. Called repeatedly
            for the assign section, and for each assign_based_on entry.
            Substitution errors are ignored at this stage, a final round of re-evaluation with ignore=False is done at the end.
            """
            # flatten assignments
            flattened = flatten_dict(assignments)
            # drop entries protected from assignment
            flattened = {name: value for name, value in flattened.items() if name not in self._protected_from_assign}
            # merge into recipe namespace
            subst.recipe._merge_(flattened)
            # perform substitutions
            try:
                flattened = evaluate_and_substitute(flattened, subst, subst.recipe, location=[whose.fqname], ignore_subst_errors=True)
            except Exception as exc:
                raise AssignmentError(f"{whose.fqname}: error evaluating assignments", exc)
            assign.update(flattened)

        # perform direct assignments
        do_assign(whose.assign)

        # add assign_based_on entries to flattened_assign
        for basevar, value_list in whose.assign_based_on.items():
            # make sure the base variable is defined
            value = None
            # it will be in subst.recipe if it was assigned, or is an input
            if basevar in subst.recipe:
                value = str(subst.recipe[basevar])
            # else it might be an input with a default, check for that
            elif basevar in self.inputs_outputs and self.inputs_outputs[basevar].default is not UNSET:
                value = str(self.inputs_outputs[basevar].default)
            # else see if it is a config setting
            else:
                comps = basevar.split('.')
                if len(comps) > 1 and comps[0] in self.config:
                    try:
                        value = self.config
                        for comp in comps:
                            value = value.get(comp)
                    except Exception as exc:
                        value = None
            # nothing found? error then
            if value is None:
                if basevar in self.inputs_outputs:
                    raise AssignmentError(f"{whose.fqname}.assign_based_on: a value for '{basevar}' was not supplied")
                elif '.' in basevar:
                    raise AssignmentError(f"{whose.fqname}.assign_based_on: '{basevar}' is not a known config item")
                else:
                    raise AssignmentError(f"{whose.fqname}.assign_based_on: '{basevar}' is not a known variable")
            # look up list of assignments
            if value not in value_list:
                if 'DEFAULT' not in value_list:
                    raise AssignmentError(f"{whose.fqname}.assign_based_on: neither the '{basevar}={value}' case nor a DEFAULT case is defined")
                value = 'DEFAULT'
            assignments = value_list.get(value)
            # an empty section maps to None, so skip 
            if assignments is None:
                continue
            if not isinstance(assignments, (dict, OrderedDict, DictConfig)):
                raise AssignmentError(f"{whose.fqname}.assign_based_on.{basevar}.{value}: mapping expected, got {type(assignments)} instead")
            # process the assignments
            do_assign(assignments)

        # do final round of substitutions
        try:
            assign = evaluate_and_substitute(assign, subst, subst.recipe, location=[whose.fqname], ignore_subst_errors=ignore_subst_errors)
        except Exception as exc:
            raise AssignmentError(f"{whose.fqname}: error evaluating assignments", exc)
        # dispatch and reassign, since substitutions may have been performed
        for key, value in assign.items():
            self.assign_value(key, value, subst=subst, whose=whose)

    def assign_value(self, key: str, value: Any, override: bool = False,
                     subst: Optional[Dict[str, Any]] = None, whose: Optional[Any] = None):
        """assigns a parameter value to the recipe. Handles nested assignments and 
        assignments to local log options.
        """
        # ignore protected assignments
        if key in self._protected_from_assign and not override:
            return

        if '.' in key:
            nesting, subkey = key.split('.', 1)
        else:
            nesting, subkey = None, key

        # assigning to input or output? Provide default            
        if key in self.inputs_outputs:
            self.log.debug(f"default params assignment: {key}={value}")
            if value is UNSET:
                if key in self.defaults:
                    del self.defaults[key]
                self.inputs_outputs[key].default = UNSET
            else:
                self.defaults[key] = value
        # assigning to a substep? Invoke nested assignment
        elif nesting is not None and nesting in self.steps:
            return self.steps[nesting].assign_value(subkey, value, override=override)
        # assigning to config?
        elif nesting == "config":
            # helper function to do nested assignment of config and subst
            def assign_nested(container, nested_key, value):
                comps = nested_key.split('.')
                while len(comps) > 1:
                    if comps[0] not in container:
                        raise AssignmentError(f"{self.fqname}: invalid assignment {key}={value}")
                    container = container[comps[0]]
                    comps.pop(0)
                container[comps[0]] = value
            assign_nested(self.config, subkey, value)
            if subst is not None and 'config' in subst:
                assign_nested(subst.config, subkey, value)
        elif nesting == "log":
            whose = whose or self
            if type(value) is Unresolved:
                self.log.debug(f"ignoring unresolved log options assignment {key}={value}")
            else:
                self.log.debug(f"log options assignment: {key}={value}")
                whose.update_log_options(**{subkey: value})
            if whose is self and subst is not None and 'recipe' in subst:
                subst.recipe.log[subkey] = value
        # in override mode, assign to assign dict for future processing
        if override:
            if value is not UNSET:
                self.assign[key] = value
            self._protected_from_assign.add(key)

    def update_log_options(self, **options):
        for setting, value in options.items():
            try:
                self.logopts[setting] = value
            except Exception as exc:
                raise AssignmentError(f"invalid {self.fqname}.log.{setting} setting", exc)
        # propagate to children
        for step in self.steps.values():
            step.update_log_options(**options)

    @property
    def finalized(self):
        return self._alias_map is not None

    def enable_step(self, label, enable=True):
        self.finalize()
        step = self.steps.get(label)
        if step is None:
            raise RecipeValidationError(f"recipe '{self.name}': unknown step {label}", log=self.log)
        if enable:
            if step._skip is True:
                self.log.warning(f"enabling step '{label}' which is normally skipped")
            elif step._skip is not False:
                self.log.warning(f"enabling step '{label}' which is normally conditionally skipped ('{step.skip}')")
            step.skip = step._skip = False
            step.skip_if_outputs = None
        else:
            self.log.warning(f"will skip step '{label}'")
            step.skip = step._skip = True

    def restrict_steps(self, tags: List[str] = [], skip_tags: List[str] = [], 
                             step_ranges: List[str] = [], enable_steps: List[str]=[]):
        try:
            # extract subsets of tags and step specifications that refer to sub-recipes
            # this will map name -> (tags, skip_tags, step_ranges, enable_steps). Name is '' for the parent recipe.
            subrecipe_entries = OrderedDict()
            def process_specifier_list(specs: List[str], num=0):
                for spec in specs:
                    if '.' in spec:
                        subrecipe, spec = spec.split('.', 1)
                        if subrecipe not in self.steps or not isinstance(self.steps[subrecipe].cargo, Recipe):
                            raise StepSelectionError(f"'{subrecipe}.{spec}' does not refer to a valid subrecipe")
                    else:
                        subrecipe = None
                    entry = subrecipe_entries.setdefault(subrecipe, ([],[],[],[])) 
                    entry[num].append(spec)
            # this builds up all the entries given on the command-line
            for num, options in enumerate((tags, skip_tags, step_ranges, enable_steps)):
                process_specifier_list(options, num)

            # process our own entries
            tags, skip_tags, step_ranges, enable_steps = subrecipe_entries.get(None, ([],[],[],[]))

            # apply enabled steps
            for name in enable_steps:
                if name in self.steps:
                    self.enable_step(name)  # config file may have skip=True, but we force-enable here
                else:
                    raise StepSelectionError(f"'{name}' does not refer to a valid step")

            # select subset based on tags/skip_tags, this will be a list of names
            tagged_steps = set()

            # if tags are given, only use steps with (tags+{"always"}-skip_tags)
            if tags:
                tags = set(tags) | {"always"}
                tags.difference_update(skip_tags)
                for step_name, step in self.steps.items():
                    if (tags & step.tags):
                        tagged_steps.add(step_name)
                self.log.info(f"{len(tagged_steps)} of {len(self.steps)} step(s) selected via tags ({', '.join(tags)})")
            # else, use steps without any tag in (skip_tags + {"never"})
            else:
                skip_tags = set(skip_tags) | {"never"}
                for step_name, step in self.steps.items():
                    if not (skip_tags & step.tags):
                        tagged_steps.add(step_name)
                if len(self.steps) != len(tagged_steps):
                    self.log.info(f"{len(self.steps) - len(tagged_steps)} step(s) skipped due to tags ({', '.join(skip_tags)})")

            # add steps explicitly enabled by --step
            if step_ranges:
                all_step_names = list(self.steps.keys())
                step_subset = set()
                for name in step_ranges:
                    if ':' in name:
                        begin, end = name.split(':', 1)
                        if begin:
                            try:
                                first = all_step_names.index(begin)
                            except ValueError as exc:
                                raise StepSelectionError(f"no such step: '{begin}' (in '{name}')")
                        else:
                            first = 0
                        if end:
                            try:
                                last = all_step_names.index(end)
                            except ValueError as exc:
                                raise StepSelectionError(f"no such step: '{end}' (in '{name}')")
                        else:
                            last = len(self.steps)-1
                        step_subset.update(name for name in all_step_names[first:last+1] if name in tagged_steps)
                    # explicit step name: enable, and add to tagged_steps
                    else:
                        if name not in all_step_names:
                            raise StepSelectionError(f"no such step: '{name}'")
                        self.enable_step(name)  # config file may have skip=True, but we force-enable here
                        step_subset.add(name)
                # specified subset becomes *the* subset
                self.log.info(f"{len(step_subset)} step(s) selected by name")
                tagged_steps = step_subset

            if not tagged_steps:
                self.log.info("no steps have been selected for execution")
                return 0
            else:
                if len(tagged_steps) != len(self.steps):
                    # apply skip flags 
                    for label, step in self.steps.items():
                        if label not in tagged_steps:
                            step.skip = step._skip = True
                # see how many steps are actually going to run
                scheduled_steps = [label for label, step in self.steps.items() if not step._skip]
                # report scheduled steps to log if (a) they're a subset or (b) any selection options were passed
                if len(scheduled_steps) != len(self.steps) or None in subrecipe_entries:
                    self.log.info(f"the following recipe steps have been selected for execution:")
                    self.log.info(f"    [bold green]{' '.join(scheduled_steps)}[/bold green]")

                # now recurse into sub-recipes
                for subrecipe, options in subrecipe_entries.items():
                    if subrecipe is not None:
                        self.steps[subrecipe].cargo.restrict_steps(*options)

                return len(scheduled_steps)
        except StepSelectionError as exc:
            log_exception(exc, log=self.log)
            raise exc


    def add_step(self, step: Step, label: str = None):
        """Adds a step to the recipe. Label is auto-generated if not supplied

        Args:
            step (Step): step object to add
            label (str, optional): step label, auto-generated if None
        """
        if self.finalized:
            raise DefinitionError("recipe '{self.name}': can't add a step to a recipe that's been finalized")

        names = [s for s in self.steps if s.cab == step.cabname]
        label = label or f"{step.cabname}_{len(names)+1}"
        self.steps[label] = step
        step.fqname = f"{self.name}.{label}"


    def add(self, cabname: str, label: str = None, 
            params: Optional[Dict[str, Any]] = None, info: str = None):
        """Add a step to a recipe. This will create a Step instance and call add_step() 

        Args:
            cabname (str): name of cab to use for this step
            label (str): Alphanumeric label (must start with a lette) for the step. If not given will be auto generated 'cabname_d' where d is the number of times a particular cab has been added to the recipe.
            params (Dict): A parameter dictionary
            info (str): Documentation of this step
        """
        return self.add_step(Step(cab=cabname, params=params, info=info), label=label)

    @dataclass
    class AliasInfo(object):
        label: str                      # step label
        step: Step                      # step
        param: str                      # parameter name
        io: Dict[str, Parameter]        # points to self.inputs or self.outputs
        from_recipe: bool = False       # if True, value propagates from recipe up to step
        from_step: bool = False         # if True, value propagates from step down to recipe

    def _add_alias(self, alias_name: str, alias_target: Union[str, Tuple], 
                    category: Optional[int] = None,
                    has_value=False):
        wildcards = False
        if type(alias_target) is str:
            # $$ maps to full name, and $ maps to last element of name
            alias_target = alias_target.replace("$$", alias_name)
            alias_target = alias_target.replace("$", alias_name.rsplit('.', 1)[-1])
            step_spec, step_param_name = alias_target.split('.', 1)
            # treat label as a "(cabtype)" specifier?
            if re.match('^\(.+\)$', step_spec):
                steps = [(label, step) for label, step in self.steps.items() 
                        if (isinstance(step.cargo, Cab) and step.cab == step_spec[1:-1]) or
                            (isinstance(step.cargo, Recipe) and step.recipe == step_spec[1:-1])]
                wildcards = True
            # treat label as a wildcard?
            elif any(ch in step_spec for ch in '*?['):
                steps = [(label, step) for label, step in self.steps.items() if fnmatch.fnmatchcase(label, step_spec)]
                wildcards = True
            # else treat label as a specific step name
            else:
                steps = [(step_spec, self.steps.get(step_spec))]
        else:
            step, step_spec, step_param_name = alias_target
            steps = [(step_spec, step)]

        for (step_label, step) in steps:
            if step is None:
                raise RecipeValidationError(f"recipe '{self.name}': alias '{alias_name}' refers to unknown step '{step_label}'", log=self.log)
            # is the alias already defined
            existing_alias = self._alias_list.get(alias_name, [None])[0]
            # find it in inputs or outputs
            input_schema = step.inputs.get(step_param_name)
            output_schema = step.outputs.get(step_param_name)
            schema = input_schema or output_schema
            # if the step was matched by a wildcard, and it doesn't have such a parameter in the schema, or else if it is
            # already explicitly specified, then we don't alias it 
            if wildcards and (schema is None or step_param_name in step.params):
                continue                    
            # no a wildcard, but parameter not defined? This is an error
            if schema is None:
                raise RecipeValidationError(f"recipe '{self.name}': alias '{alias_name}' refers to unknown step parameter '{step_label}.{step_param_name}'", log=self.log)
            # implicit inputs cannot be aliased
            if input_schema and input_schema.implicit:
                raise RecipeValidationError(f"recipe '{self.name}': alias '{alias_name}' refers to implicit input '{step_label}.{step_param_name}'", log=self.log)
            # if alias is already defined, check for conflicts
            if existing_alias is not None:
                io = existing_alias.io
                if io is self.outputs:
                    raise RecipeValidationError(f"recipe '{self.name}': output alias '{alias_name}' is defined more than once", log=self.log)
                elif output_schema:
                    raise RecipeValidationError(f"recipe '{self.name}': alias '{alias_name}' refers to both an input and an output", log=self.log)
                alias_schema = io[alias_name] 
                # now we know it's a multiply-defined input, check for type consistency
                if alias_schema.dtype != schema.dtype:
                    raise RecipeValidationError(f"recipe '{self.name}': alias '{alias_name}': dtype {schema.dtype} of '{step_label}.{step_param_name}' doesn't match previous dtype {alias_schema.dtype}", log=self.log)
                orig_schema = self._orig_alias_schema[alias_name]
            # else alias not yet defined, insert a schema
            else:
                # get recipe's original schema for the parameter
                io = self.inputs if input_schema else self.outputs
                # if we have a schema defined for the alias, some params must be inherited from it 
                orig_schema = io.get(alias_name)
                self._orig_alias_schema[alias_name] = orig_schema
                # define schema based on copy of the target, but preserve default
                io[alias_name] = copy.copy(schema)
                alias_schema = io[alias_name] 
                # check that 
                # default set from own schema, ignoring parameter setting 
                if orig_schema is not None and orig_schema.default is not UNSET:
                    # also clear parameter setting to propagate our default
                    if step_param_name in step.params:
                        del step.params[step_param_name]
                    alias_schema.default = orig_schema.default
                if orig_schema and orig_schema.info:
                    alias_schema.info = orig_schema.info
                # required flag overrides, if set from our own schema
                if orig_schema is not None and orig_schema.required is not None:
                    alias_schema.required = orig_schema.required
                # category is set by argument, else from own schema, else from target
                if category is not None:
                    alias_schema.category = category
                elif orig_schema is not None and orig_schema.category is not None:
                    alias_schema.category = orig_schema.category
                # parameter is not required if alias target is set in step
                if step_param_name in step.params or \
                        step_param_name in step.cargo.defaults or \
                        schema.default is not UNSET:
                    alias_schema.required = False
                    alias_schema.default = UNSET
                    # mark it as hidden -- no need to expose parameters that are internally set this way
                    alias_schema.category = ParameterCategory.Hidden

            # if step parameter is implicit, mark the alias as implicit. Note that this only applies to outputs
            if schema.implicit:
                alias_schema.implicit = Unresolved(f"{step_label}.{step_param_name}")   # will be resolved when propagated from step
                self._implicit_params.add(alias_name)

            # this is True if the step's parameter is defined in any way (set, default, or implicit)
            have_step_param = step_param_name in step.params or step_param_name in step.cargo.defaults or \
                alias_schema.default is not UNSET or alias_schema.implicit is not None

            # if the step parameter is set and ours isn't, mark our schema as having a default
            if have_step_param and alias_schema.default is UNSET:
                alias_schema.default = DeferredAlias(f"{step_label}.{step_param_name}")

            # alias becomes required if any step parameter it refers to is required and not set, unless
            # the original schema forces a required value
            if schema.required and not have_step_param and (orig_schema is None or orig_schema.required is None):
                alias_schema.required = True

            self._alias_map[step_label, step_param_name] = alias_name, orig_schema
            self._alias_list.setdefault(alias_name, []).append(Recipe.AliasInfo(step_label, step, step_param_name, io))

    def finalize(self, config=None, log=None, name=None, fqname=None, nesting=0):
        if not self.finalized:
            config = config or stimela.CONFIG

            # fully qualified name, i.e. recipe_name.step_name.step_name etc.
            self.fqname = fqname = fqname or self.fqname or self.name

            # if logger is not provided, then init one
            if log is None:
                log = stimela.logger().getChild(self.fqname)
                log.propagate = True

            # init and/or update logger options
            self.logopts = config.opts.log.copy()

            # update file logger
            logsubst = SubstitutionNS(config=config, info=dict(fqname=fqname))
            stimelogging.update_file_logger(log, self.logopts, nesting=nesting, subst=logsubst, location=[self.fqname])

            # call Cargo's finalize method
            super().finalize(config, log=log, fqname=fqname, nesting=nesting)

            # finalize steps
            for label, step in self.steps.items():
                step_log = log.getChild(label)
                step_log.propagate = False
                try:
                    step.finalize(config, log=step_log, fqname=f"{fqname}.{label}", nesting=nesting+1)
                    # check that per-step assignments don't clash with i/o parameters
                    step.assign = self.flatten_param_dict(OrderedDict(), step.assign)
                    self.validate_assignments(step.assign, step.assign_based_on, f"{fqname}.{label}")
                except Exception as exc:
                    raise StepValidationError(f"error validating step '{label}'", exc, 
                                tb=not isinstance(exc, ScabhaBaseException))

            # collect aliases
            self._alias_map = OrderedDict()
            self._alias_list = OrderedDict()
            self._orig_alias_schema = OrderedDict()

            # collect from inputs and outputs
            for io in self.inputs, self.outputs:
                for name, schema in io.items():
                    if schema.aliases:
                        ## NB skip this check, allow aliases to override
                        # if schema.dtype != "str" or schema.choices or schema.writable:
                        #     raise RecipeValidationError(f"recipe '{self.name}': alias '{name}' should not specify type, choices or writability", log=log)
                        for alias_target in schema.aliases:
                            self._add_alias(name, alias_target)

            # collect from aliases section
            for name, alias_list in self.aliases.items():
                for alias_target in alias_list:
                    self._add_alias(name, alias_target)

            # automatically make aliases for step parameters that are unset, and don't have a default, and aren't implict 
            for label, step in self.steps.items():
                for name, schema in step.inputs_outputs.items():
                    # does it have a value set
                    has_value = name in step.params or name in step.cargo.defaults or \
                                schema.default is not UNSET 
                    if (label, name) not in self._alias_map and not schema.implicit and not has_value:
                        auto_name = f"{label}.{name}"
                        if auto_name in self.inputs or auto_name in self.outputs:
                            raise RecipeValidationError(f"recipe '{self.name}': auto-generated parameter name '{auto_name}' conflicts with another name. Please define an explicit alias for this.", log=log)
                        self._add_alias(auto_name, (step, label, name), 
                                        category=ParameterCategory.Required if schema.required and not has_value  
                                        else ParameterCategory.Obscure)

            # these will be re-merged when needed again
            self._inputs_outputs = None

            # check that for-loop is valid, if defined
            if self.for_loop is not None:
                # if for_loop.over is a str, treat it as a required input
                if type(self.for_loop.over) is str:
                    if self.for_loop.over not in self.inputs:
                        raise RecipeValidationError(f"recipe '{self.name}': for_loop.over={self.for_loop.over} is not a defined input", log=log)
                    # this becomes a required input
                    self.inputs[self.for_loop.over].required = True
                # else treat it as a list of values to be iterated over (and set over=None to indicate this)
                elif type(self.for_loop.over) in (list, tuple, ListConfig):
                    self._for_loop_values = list(self.for_loop.over)
                    self.for_loop.over = None
                else:
                    raise RecipeValidationError(f"recipe '{self.name}': for_loop.over is of invalid type {type(self.for_loop.over)}", log=log)

                # # insert empty loop variable
                # if self.for_loop.var not in self.assign:
                #     self.assign[self.for_loop.var] = ""

    def _prep_step(self, label, step, subst):
        parts = label.split("-")
        info = subst.info
        info.fqname = f"{self.fqname}.{label}"
        info.label = label 
        info.label_parts = parts
        info.suffix = parts[-1] if len(parts) > 1 else ''
        subst.current = step.params
        subst.steps[label] = subst.current

    def _preprocess_parameters(self, params: Dict[str, Any]):
        # split parameters into our own, and per-step, and UNSET directives
        own_params = {}
        unset_params = set()
        for name, value in params.items():
            # if self.for_loop is not None and name == self.for_loop.var:
            #     continue # unset_params.add(name)
            if name in self.inputs_outputs:
                # if value == "UNSET":
                #     unset_params.add(name)
                # elif value == "EMPTY":
                #     own_params[name] = ""
                if value is UNSET: 
                    unset_params.add(name)
                else:
                    own_params[name] = value
            elif '.' not in name: 
                raise ParameterValidationError(f"'{name}' does not refer to a known parameter")
            else:
                label, subname = name.split('.', 1)
                substep = self.steps.get(label)
                if substep is None:
                    raise ParameterValidationError(f"'{name}' does not refer to a known parameter or a step")
                substep.params[subname] = value
        return own_params, unset_params


    def prevalidate(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, root=False):
        self.finalize()
        self.log.debug("prevalidating recipe")
        errors = []

        # split parameters into our own, and per-step, and UNSET directives
        params,  unset_params = self._preprocess_parameters(params)

        subst_outer = subst  # outer dictionary is used to prevalidate our parameters

        subst = SubstitutionNS()
        info = SubstitutionNS(fqname=self.fqname, label='', label_parts=[], suffix='')
        # mutable=False means these sub-namespaces are not subject to {}-substitutions
        subst._add_('info', info.copy(), nosubst=True)
        subst._add_('config', self.config, nosubst=True) 
        subst._add_('steps', {}, nosubst=True)
        subst._add_('previous', {}, nosubst=True)
        subst.recipe = SubstitutionNS(**params)
        subst.recipe.log = self.logopts
        if root:
            subst.root = subst.recipe

        if subst_outer is not None:
            if 'root' in subst_outer:
                subst._add_('root', subst_outer.root, nosubst=True)
            if 'recipe' in subst_outer:
                subst._add_('parent', subst_outer.recipe, nosubst=True)
        else:
            subst_outer = SubstitutionNS()
            subst_outer._add_('info', info.copy(), nosubst=True)
            subst_outer._add_('config', self.config, nosubst=True) 
            subst_outer.current = subst.recipe

        # update assignments
        self.update_assignments(subst, params=params, ignore_subst_errors=True)
        # this may have changed the file logger, so update
        stimelogging.update_file_logger(self.log, self.logopts, nesting=self.nesting, subst=subst, location=[self.fqname])

        # add for-loop variable to inputs, if expected there
        if self.for_loop is not None and self.for_loop.var in self.inputs:
            params[self.for_loop.var] = Placeholder(self.for_loop.var)

        # prevalidate our own parameters. This substitutes in defaults and does {}-substitutions
        # we call this twice, potentially, so define as a function
        def prevalidate_self(params):
            try:
                params1 = Cargo.prevalidate(self, params, subst=subst_outer)
                # mark params that have become unset 
                unset_params.update(set(params) - set(params1))
                params = params1
                # validate for-loop, if needed
                self.validate_for_loop(params, strict=False)

            except ScabhaBaseException as exc:
                errors.append(exc)
            except Exception as exc:
                errors.append(RecipeValidationError("recipe failed prevalidation", exc, tb=True))

            # merge again, since values may have changed
            subst.recipe._merge_(params)
            return params

        params = prevalidate_self(params)

        # propagate alias values up to substeps, except for implicit values (these only ever propagate down to us)
        for name, aliases in self._alias_list.items():
            if name in params and type(params[name]) is not DeferredAlias and name not in self._implicit_params:
                for alias in aliases:
                    alias.from_recipe = True
                    alias.step.update_parameter(alias.param, params[name])
            elif name in unset_params:
                for alias in aliases:
                    alias.from_recipe = True
                    alias.step.unset_parameter(alias.param)

        # prevalidate step parameters 
        # we call this twice, potentially, so define as a function

        def prevalidate_steps():
            for label, step in self.steps.items():
                self._prep_step(label, step, subst)
                # update assignments, since substitutions (info.fqname and such) may have changed
                self.update_assignments(subst, params=params, ignore_subst_errors=True)
                # update assignments based on step content
                self.update_assignments(subst, whose=step, params=params, ignore_subst_errors=True)

                try:
                    step_params = step.prevalidate(subst)
                    subst.current._merge_(step_params)   # these may have changed in prevalidation
                except ScabhaBaseException as exc:
                    errors.append(RecipeValidationError(f"step '{label}' failed prevalidation", exc))
                except Exception as exc:
                    errors.append(RecipeValidationError(f"step '{label}' failed prevalidation", exc, tb=True))

                # revert to recipe-level assignments
                self.update_assignments(subst, params=params, ignore_subst_errors=True)
                subst.previous = subst.current
                subst.steps[label] = subst.previous

        prevalidate_steps()

        # now check for aliases that need to be propagated up/down
        if not errors:
            revalidate_self = revalidate_steps = False
            for name, aliases in self._alias_list.items():
                # propagate up if alias is not set, or it is implicit=Unresolved (meaning it gets set from an implicit substep parameter)
                if name not in params or type(params[name]) is DeferredAlias or type(self.inputs_outputs[name].implicit) is Unresolved:
                    from_step = False
                    for alias in aliases:
                        # if alias is set in step but not with us, mark it as propagating down
                        if alias.param in alias.step.validated_params:
                            alias.from_step = from_step = revalidate_self = True
                            params[name] = alias.step.validated_params[alias.param]
                            # and break out, we do this for the first matching step only
                            break
                    # if we propagated an input value down from a step, check if we need to propagate it up to any other steps
                    # note that this only ever applies to inputs
                    if from_step:
                        for alias in aliases:
                            if not alias.from_step:
                                alias.from_recipe = revalidate_steps = True
                                alias.step.update_parameter(alias.param, params[name])

            # do we or any steps need to be revalidated?
            if revalidate_self:
                params = prevalidate_self(params)
            if revalidate_steps:
                prevalidate_steps()

        # check for missing parameters
        missing_params = [name for name, schema in self.inputs_outputs.items() if schema.required and name not in params]
        if missing_params:
            n = len(missing_params)
            msg = f"""recipe is missing {n} required parameter{'s' if n>1 else ''}:"""
            errors.append(RecipeValidationError(msg, missing_params))

        if errors:
            if len(errors) == 1:
                raise errors[0]
            else:
                raise RecipeValidationError(f"recipe '{self.name}': {len(errors)} errors", errors)

        self._prevalidated_steps = subst.steps

        self.log.debug("recipe pre-validated")

        return params

    def validate_for_loop(self, params, strict=False):
        # in case of for loops, get list of values to be iterated over 
        if self.for_loop is not None:
            # get scatter value
            if 'for_loop.scatter' in params:
                scatter = params['for_loop.scatter']
            elif 'for_loop.over' in self.assign:
                scatter = self.assign['for_loop.scatter']
            else:
                scatter = self.for_loop.scatter
            if type(scatter) is bool:
                scatter = -1 if scatter else 0
            elif type(scatter) is not int:
                raise ParameterValidationError(f"for_loop.scattter={scatter}: bool or int expected")
            self._for_loop_scatter = scatter

            # the over list can be in the for_loop clause, or in inputs
            if 'for_loop.over' in params:
                values = params['for_loop.over']
            elif 'for_loop.over' in self.assign:
                values = self.assign['for_loop.over']
            elif self.for_loop.over is not None:
                # check that it's legal
                if self.for_loop.over in self.assign:
                    values = self.assign[self.for_loop.over]
                elif self.for_loop.over in params:
                    values = params[self.for_loop.over]
                elif self.for_loop.over not in self.inputs:
                    raise ParameterValidationError(f"recipe '{self.name}': for_loop.over={self.for_loop.over} does not refer to a known parameter")
                else:
                    raise ParameterValidationError(f"recipe '{self.name}': for_loop.over={self.for_loop.over} is unset")
                if strict and isinstance(values, Unresolved):
                    raise ParameterValidationError(f"recipe '{self.name}': for_loop.over={self.for_loop.over} is unresolved")
            else:
                if self._for_loop_values is None:
                    raise ParameterValidationError(f"recipe '{self.name}': for_loop.over is unset")
                values = self._for_loop_values 
            # finalize list of values
            if type(values) is ListConfig:
                values = list(values)
            elif not isinstance(values, (list, tuple)):
                values = [values]
            if self._for_loop_values is None:
                self.log.debug(f"recipe is a for-loop with '{self.for_loop.var}' iterating over {len(values)} values")
                self.log.debug(f"loop values: {values}")
            self._for_loop_values = values
        # else fake a single-value list
        else:
            self._for_loop_values = [None]

    def validate_inputs(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, loosely=False):

        params, _ = self._preprocess_parameters(params)

        if subst is None:
            subst = SubstitutionNS()
            info = SubstitutionNS(fqname=self.fqname)
            subst._add_('info', info, nosubst=True)
            subst._add_('config', self.config, nosubst=True) 

            subst.recipe = SubstitutionNS(**params)
            subst.current = subst.recipe

        if 'current' in subst:
            subst.current._add_('steps', self._prevalidated_steps, nosubst=True)
        
        self.update_assignments(subst, params=params, ignore_subst_errors=True)

        params = Cargo.validate_inputs(self, params, subst=subst, loosely=loosely)

        self.validate_for_loop(params, strict=True)

        # # in case of a for-loop, assign first iterant
        # if self.for_loop is not None:
        #     if self.for_loop.var in self.inputs:
        #         params[self.for_loop.var] = IterantPlaceholder(self.for_loop.var)
        #     else:
        #         self.assign[self.for_loop.var] = IterantPlaceholder(self.for_loop.var)

        return params

    ## NB: OMS: is this really used or needed anywhere?
    # def _link_steps(self):
    #     """
    #     Adds  next_step and previous_step attributes to the recipe. 
    #     """
    #     steps = list(self.steps.values())
    #     N = len(steps)
    #     # Nothing to link if only one step
    #     if N == 1:
    #         return

    #     for i in range(N):
    #         step = steps[i]
    #         if i == 0:
    #             step.next_step = steps[1]
    #             step.previous_step = None
    #         elif i > 0 and i < N-2:
    #             step.next_step = steps[i+1]
    #             step.previous_step = steps[i-1]
    #         elif i == N-1:
    #             step.next_step = None
    #             step.previous_step = steps[i-2]

    def summary(self, params: Dict[str, Any], recursive=True, ignore_missing=False):
        """Returns list of lines with a summary of the recipe state
        """
        lines = [f"recipe '{self.name}':"] + [f"  {name} = {value}" for name, value in params.items()]
        if not ignore_missing:
            lines += [f"  {name} = ???" for name in self.inputs_outputs if name not in params]
        if recursive:
            lines.append("  steps:")
            for name, step in self.steps.items():
                stepsum = step.summary()
                lines.append(f"    {name}: {stepsum[0]}")
                lines += [f"    {x}" for x in stepsum[1:]]
        return lines

    _root_recipe_ns = None

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        Cargo.rich_help(self, tree, max_category=max_category)
        if self.for_loop:
            loop_tree = tree.add("For loop:")
            if self._for_loop_values is not None:
                over = f"{len(self._for_loop_values)} values"
            else:
                over = f"[bold]{self.for_loop.over}[/bold]"
            loop_tree.add(f"iterating [bold]{self.for_loop.var}[/bold] over {over}")
        if self.steps:
            have_skips = any(step._skip for step in self.steps.values())
            steps_tree = tree.add(f"Steps (note [italic]some steps[/italic] are skipped by default):" 
                                if have_skips else "Steps:")
            table = rich.table.Table.grid("", "", "", padding=(0,2)) # , show_header=False, show_lines=False, box=rich.box.SIMPLE)
            steps_tree.add(table)            
            for label, step in self.steps.items():
                style = "italic" if step._skip else "bold"
                table.add_row(f"[{style}]{label}[/{style}]", step.info)
                if step.tags:
                    table.add_row("", f"[italic](tags: {', '.join(step.tags)})[/italic]")
        else:
            steps_tree = tree.add("No recipe steps defined")


    def _update_aliases(self, name: str, value: Any):
        """Propagates recipe aliases up top parameters

        Args:
            name (str): name of recipe parameter
            value (Any): value
        """
        for alias in self._alias_list.get(name, []):
            if alias.from_recipe:
                alias.step.update_parameter(alias.param, value)


    def _iterate_loop_worker(self, params, info, subst, count, iter_var, raise_exc=True):
        """"
        Needed for concurrency
        """
        outputs = {}
        exception = tb = None
        try:
            # if for-loop, assign new value
            if self.for_loop:
                self.log.info(f"for loop iteration {count}: {self.for_loop.var} = {iter_var}")
                print(f"for loop iteration {count}: {self.for_loop.var} = {iter_var}")
                if self.for_loop.var in self.inputs_outputs:
                    params[self.for_loop.var] = iter_var
                else:
                    self.assign[self.for_loop.var] = iter_var
                # update variable index
                self.assign[f"{self.for_loop.var}@index"] = count
                # update alias
                self._update_aliases(self.for_loop.var, iter_var)
                # update status display
                status = None
                status_dict = dict(index0=count, 
                            index1=count+1, total=len(self._for_loop_values),
                            var=self.for_loop.var, value=iter_var)
                if self.for_loop.display_status:
                    try:
                        status = self.for_loop.display_status.format(**status_dict)
                    except Exception as exc:
                        self.log.warning(f"error formatting for-loop status: {exc}, falling back on default status display")
                if status is None:
                    status = "{index1}/{total}".format(**status_dict)
                stimelogging.declare_subtask_attributes(status)

            for label, step in self.steps.items():
                # update step info
                self._prep_step(label, step, subst)
                # reevaluate recipe level assignments (info.fqname etc. have changed)
                self.update_assignments(subst, params=params)
                # evaluate step-level assignments
                self.update_assignments(subst, whose=step, params=params)
                # step logger may have changed
                stimelogging.update_file_logger(step.log, step.logopts, nesting=step.nesting, subst=subst, location=[step.fqname])
                # set our info back temporarily to update log assignments
                info_step = subst.info
                subst.info = info.copy()
                subst.info = info_step

                if step.skip is True:
                    self.log.debug(f"step '{label}' will be explicitly skipped")
                else:
                    self.log.info(f"processing step '{label}'")
                    if step.info:
                        self.log.info(f"  ({step.info})", extra=dict(color="GREEN", boldface=True))
                try:
                    #step_params = step.run(subst=subst.copy(), batch=batch)  # make a copy of the subst dict since recipe might modify
                    step_params = step.run(subst=subst.copy(), parent_log=self.log)  # make a copy of the subst dict since recipe might modify
                except ScabhaBaseException as exc:
                    newexc = StimelaStepExecutionError(f"step '{step.fqname}' has failed, aborting the recipe", exc)
                    if not exc.logged:
                        log_exception(newexc, log=step.log)
                    raise newexc

                # put step parameters into previous and steps[label] again, as they may have changed based on outputs)
                subst.previous = step_params
                subst.steps[label] = subst.previous
                # revert to recipe level assignments

                # now check for output aliases that need to be propagated down from steps
                self.update_assignments(subst, whose=self, params=params)
                for name, aliases in self._alias_list.items():
                    for alias in aliases:
                        if alias.from_step:
                            # if step was skipped, mark output as not required
                            if alias.step._skip:
                                self.outputs[name].required = False
                            # if step output is validated, add it to our output 
                            if alias.param in alias.step.validated_params:
                                outputs[name] = alias.step.validated_params[alias.param]

        except Exception as exc:
            # raise exception up if asked to
            if raise_exc:
                raise
            # else will be returned
            exception = exc
            tb = FormattedTraceback(sys.exc_info()[2])

        return task_stats.collect_stats(), outputs, exception, tb

    def _run(self, params, subst=None) -> Dict[str, Any]:
        """Internal recipe run method. Meant to be called from a wrapper Step object (which validates the parameters, etc.)

        Parameters
        ----------

        Returns
        -------
        Dict[str, Any]
            Dictionary of formal outputs

        Raises
        ------
        RecipeValidationError
        """

        # set up substitution namespace
        subst_outer = subst
        if subst is None:
            subst = SubstitutionNS()

        info = SubstitutionNS(fqname=self.fqname, label='', label_parts=[], suffix='')
        # nosubst=True means these sub-namespaces are not subject to {}-substitutions
        subst._add_('info', info.copy(), nosubst=True)
        subst._add_('config', self.config, nosubst=True)
        subst._add_('steps', {}, nosubst=True)
        subst._add_('previous', {}, nosubst=True)
        subst._add_('current', {}, nosubst=True)
            
        subst.recipe = SubstitutionNS(**params)
        subst.recipe.log = self.logopts
        subst.recipe._add_('steps', subst.steps, nosubst=True)

        if subst_outer is not None:
            if 'root' in subst_outer:
                subst._add_('root', subst_outer.root, nosubst=True)
            if 'recipe' in subst_outer:
                subst._add_('parent', subst_outer.recipe, nosubst=True)
        else:
            subst.root = subst.recipe

        subst_copy = subst.copy()

        try:
            # # update variable assignments
            # self.update_assignments(subst, params=params)
            # # log options may have changed, so adjust
            # stimelogging.update_file_logger(self.log, self.logopts, nesting=self.nesting, subst=subst, location=[self.fqname])

            # # Harmonise before running
            # self._link_steps()

            self.log.info(f"running recipe '{self.name}'")

            # our inputs have been validated, so propagate aliases to steps. Check for missing stuff just in case
            for name, schema in self.inputs.items():
                if name in params:
                    value = params[name]
                    if isinstance(value, Unresolved) and not isinstance(value, Placeholder):
                        raise RecipeValidationError(f"recipe '{self.name}' has unresolved input '{name}'", log=self.log)
                    self._update_aliases(name, value)
                elif schema.required and (self.for_loop is None or name != self.for_loop.var): 
                        raise RecipeValidationError(f"recipe '{self.name}' is missing required input '{name}'", log=self.log)

            # form list of arguments for each invocation of the loop worker
            loop_worker_args = []
            for count, iter_var in enumerate(self._for_loop_values):
                loop_worker_args.append((params, info, subst, count, iter_var))

            # if scatter is enabled, use a process pool
            if self._for_loop_scatter:
                num_workers = self._for_loop_scatter if self._for_loop_scatter > 0 else len(loop_worker_args)
                with ProcessPoolExecutor(num_workers) as pool:
                    # submit each iterant to pool
                    futures = [pool.submit(self._iterate_loop_worker, *args, raise_exc=False) for args in loop_worker_args]
                    stats = [f.result() for f in futures]
                    # update task stats, since they're recorded independently within each step, as well
                    # as get any exceptions from the nesting
                    errors = []
                    nfail = 0
                    for f in futures:
                        stats, outputs, exc, tb = f.result()
                        task_stats.add_missing_stats(stats)
                        if exc is not None:
                            errors.append(exc)
                            if not isinstance(exc, ScabhaBaseException):
                                errors.append(tb)
                            nfail += 1
                    if errors:
                        pool.shutdown()
                        raise StimelaRuntimeError(f"{nfail}/{len(loop_worker_args)} loop iterations failed", errors)
            # else just iterate directly
            else:
                for args in loop_worker_args:
                    _, outputs, _, _ = self._iterate_loop_worker(*args, raise_exc=True) 
            
            # either way, outputs contains output aliases from the last iteration
            params.update(**outputs)

            # current namespace becomes recipe again
            subst.current = subst.recipe
            
            # # evaluate implicit outputs
            # implicits = {name: schema.implcit for name, schema in self.outputs if schema.implicit}
            # params.update(**evaluate_and_substitute(implicits, subst, subst.current, location=self.fqname))

            self.log.info(f"recipe '{self.name}' executed successfully")
            return OrderedDict((name, value) for name, value in params.items() if name in self.outputs)
        finally:
            steps = subst.steps
            subst.update(subst_copy)
            subst.current.steps = steps


    # def run(self, **params) -> Dict[str, Any]:
    #     """Public interface for running a step. Keywords are passed in as step parameters

    #     Returns
    #     -------
    #     Dict[str, Any]
    #         Dictionary of formal outputs
    #     """
    #     return Step(recipe=self, params=params, info=f"wrapper step for recipe '{self.name}'").run()

StepSchema = OmegaConf.structured(Step)
RecipeSchema = OmegaConf.structured(Recipe)

class PyRecipe(Recipe):
    """ 
        Interface to Recipe class for python recipes (not YAML recipes)
    """
    def __init__(self, name, dirs, backend=None, info=None, log=None):

        self.backend = backend
        self.name = name

        self.inputs: Dict[str, Any] = {}

        for dir_item in dirs:
            self.inputs[dir_item] = { 
                "dtype": Directory,
                "default": dirs[dir_item]
            }




from scabha.cargo import Parameter
from typing import Dict, Any


_jones_schema = dict(
    t_int=Parameter(dtype="Union[str, int]", info="time solution interval"),
    f_int=Parameter(dtype="Union[str, int]", info="frequency solution interval"),
) 


def cubical_schema(params: Dict[str, Any], inputs: Dict[str, Parameter], outputs: Dict[str, Parameter]):
    inputs = inputs.copy()

    for jones in params['jones']:
        for key, value in _jones_schema.items():
            inputs[f"{jones}.{key}"] = value

    return inputs, outputs



import sys
import os.path
import pytest
from scabha import configuratt
from scabha.configuratt import ConfigurattError
from omegaconf import OmegaConf
from typing import *

#testdir = os.path.dirname(os.path.abspath(__file__))

# Change into directory where test_recipy.py lives
# As suggested by https://stackoverflow.com/questions/62044541/change-pytest-working-directory-to-test-case-directory
@pytest.fixture(autouse=True)
def change_test_dir(request, monkeypatch):
    monkeypatch.chdir(request.fspath.dirname)

def test_includes():
    path = "testconf.yaml"
    conf, deps = configuratt.load(path, use_sources=[], verbose=True, use_cache=False)
    assert conf.x.y2.z1 == 1

    missing = configuratt.check_requirements(conf, [], strict=False)
    assert len(missing) == 2                   # 2 failed reqs
    assert 'yy' not in conf.requirements2.x    # this one was contingent, so section was deleted
    assert missing[0][2] is None               # this one was contingent, so no error
    assert type(missing[1][2]) is ConfigurattError  # this one was required, so yes error

    try:
        conf, deps = configuratt.load(path, use_sources=[], verbose=True, use_cache=False)
        missing = configuratt.check_requirements(conf, [], strict=True)
        raise RuntimeError("Error was expected here!")
    except configuratt.ConfigurattError as exc:
        print(f"Exception as expected: {exc}")

    nested = ["test_nest_a.yml", "test_nest_b.yml", "test_nest_c.yml"]
    nested = [os.path.join(os.path.dirname(path), name) for name in nested]

    conf1, deps1 = configuratt.load_nested(nested, 
                                            typeinfo=Dict[str, Any], nameattr="_name", verbose=True, use_cache=False)
    conf['nested'] = conf1
    OmegaConf.save(conf, sys.stderr)

    deps.update(deps1)

    print(f"Dependencies are: {deps.get_description()}")


if __name__ == "__main__":
    test_includes(sys.argv[1] if len(sys.argv) > 1 else None)


import fnmatch
import os, sys
import click
from typing import *
from rich.tree import Tree
from rich.table import Table
from rich import box
from rich import print as rich_print
from omegaconf import OmegaConf

import stimela
from scabha import configuratt
from stimela import logger, log_exception
from stimela.main import cli
from scabha.cargo import ParameterCategory
from stimela.kitchen.recipe import Recipe
from stimela.kitchen.cab import Cab
from stimela.config import ConfigExceptionTypes
from stimela.exceptions import RecipeValidationError

from .run import load_recipe_file

@cli.command("help",
    help="""
    Print help on a cab or a recipe.
    """)
@click.option("do_list", "-l", "--list", is_flag=True,
                help="""Lists the available cabs and recipes, including custom-defined ones.""")
@click.option("-I", "--implicit", is_flag=True,
                help="""Increases level of detail to include implicit inputs/outputs.""")
@click.option("-O", "--obscure", is_flag=True,
                help="""Increases level of detail to include implicit and obscure inputs/outputs.""")
@click.option("-A", "--all", is_flag=True,
                help="""Increases level of detail to include all inputs/outputs.""")
@click.option("-R", "--required", is_flag=True,
                help="""Decreases level of detail to include required inputs/outputs only.""")
@click.argument("items", nargs=-1, metavar="filename.yml|cab name|recipe name|...") 
def help(items: List[str] = [], do_list=False, implicit=False, obscure=False, all=False, required=False):

    log = logger()
    top_tree = Tree(f"stimela help {' '.join(items)}", guide_style="dim")
    found_something = False
    default_recipe = None

    if required:
        max_category = ParameterCategory.Required
    else:
        max_category = ParameterCategory.Optional
    if all:
        max_category = ParameterCategory.Hidden
    elif obscure:
        max_category = ParameterCategory.Obscure
    elif implicit:
        max_category = ParameterCategory.Implicit
        
    def load_recipe(name: str, section: Dict):
        try:
            if not section.get('name'):
                section.name = name
            recipe = Recipe(**section)
            recipe.finalize(fqname=name)
            return recipe
        except Exception as exc:
            if not isinstance(exc, RecipeValidationError):
                exc = RecipeValidationError(f"error loading recipe '{name}'", exc)
            log_exception(exc)
            sys.exit(2)


    for item in items:
        # a filename -- treat it as a config
        if os.path.isfile(item):
            log.info(f"loading recipe/config {item}")

            # if file contains a recipe entry, treat it as a full config (that can include cabs etc.)
            conf, recipe_deps = load_recipe_file(item)

            # anything that is not a standard config section will be treated as a recipe
            recipes = [name for name in conf if name not in stimela.CONFIG]

            if len(recipes) == 1:
                default_recipe = recipes[0]

            for name in recipes:
                try:
                    # cast section to Recipe and remove from loaded conf
                    recipe = OmegaConf.structured(Recipe)
                    recipe = OmegaConf.unsafe_merge(recipe, conf[name])
                except Exception as exc:
                    log.error(f"recipe '{name}': {exc}")
                    sys.exit(2)
                del conf[name]
                # add to global namespace
                stimela.CONFIG.lib.recipes[name] = recipe

            # the rest is safe to merge into config as is
            stimela.CONFIG = OmegaConf.unsafe_merge(stimela.CONFIG, conf)
        
        # else treat as a wildcard for recipe names or cab names
        else:
            recipe_names = fnmatch.filter(stimela.CONFIG.lib.recipes.keys(), item)
            cab_names = fnmatch.filter(stimela.CONFIG.cabs.keys(), item)
            if not recipe_names and not cab_names:
                log.error(f"'{item}' does not match any files, recipes or cab names. Try -l/--list")
                sys.exit(2)

            for name in recipe_names:
                recipe = load_recipe(name, stimela.CONFIG.lib.recipes[name])
                tree = top_tree.add(f"Recipe: [bold]{name}[/bold]")
                recipe.rich_help(tree, max_category=max_category)
            
            for name in cab_names:
                cab = Cab(**stimela.CONFIG.cabs[name])
                cab.finalize(config=stimela.CONFIG)
                tree = top_tree.add(f"Cab: [bold]{name}[/bold]")
                cab.rich_help(tree, max_category=max_category)

            found_something = True

    if do_list or (not found_something and not default_recipe):

        if stimela.CONFIG.lib.recipes:
            subtree = top_tree.add("Recipes:")
            table = Table.grid("", "", padding=(0,2))
            for name, recipe in stimela.CONFIG.lib.recipes.items():
                table.add_row(f"[bold]{name}[/bold]", recipe.info)
            subtree.add(table)
        elif not do_list and not found_something:
            log.error(f"nothing particular to help on, perhaps specify a recipe name or a cab name, or use -l/--list")
            sys.exit(2)

    if default_recipe and not found_something:
        recipe = load_recipe(name, stimela.CONFIG.lib.recipes[default_recipe])
        tree = top_tree.add(f"Recipe: [bold]{default_recipe}[/bold]")
        recipe.rich_help(tree, max_category=max_category)

    if do_list:
        subtree = top_tree.add("Cabs:")
        table = Table.grid("", "", padding=(0,2))
        for name, cab in stimela.CONFIG.cabs.items():
            table.add_row(f"[bold]{name}[/bold]", cab.info)
        subtree.add(table)            
        
    rich_print(top_tree)





import sys
import  os.path
import  re
import logging
import traceback
import copy
from types import TracebackType
from typing import Optional, OrderedDict, Union
from omegaconf import DictConfig
from scabha.exceptions import ScabhaBaseException, FormattedTraceback
from scabha.substitutions import SubstitutionNS, forgiving_substitutions_from
import rich.progress
import rich.logging
from rich.tree import Tree
from rich import print as rich_print
from rich.markup import escape
from rich.padding import Padding

from . import task_stats
from .task_stats import declare_subtask, declare_subtask_attributes, \
                        declare_subcommand, update_process_status, \
                        run_process_status_update

class FunkyMessage(object):
    """Class representing a message with two versions: funky (with markup), and boring (no markup)"""
    def __init__(self, funky, boring=None, prefix=None):
        self.funky = funky
        self.boring = boring if boring is not None else funky
        self.prefix = prefix
    def __str__(self):
        return self.funky
    def __add__(self, other):
        if isinstance(other, FunkyMessage):
            return FunkyMessage(f"{self}{other}", f"{self.boring}{other.boring}")
        else:
            return FunkyMessage(f"{self}{other}", f"{self.boring}{other}")

def defunkify(arg: Union[str, FunkyMessage]):
    return arg.boring if isinstance(arg, FunkyMessage) else arg

class MultiplexingHandler(logging.Handler):
    """handler to send INFO and below to stdout, everything above to stderr"""
    def __init__(self, info_stream=sys.stdout, err_stream=sys.stderr):
        super(MultiplexingHandler, self).__init__()
        self.info_handler = logging.StreamHandler(info_stream)
        self.err_handler = logging.StreamHandler(err_stream)
        self.multiplex = True

    def emit(self, record):
        # does record come with its own handler? Rather use that
        if hasattr(record, 'custom_console_handler'):
            handler = record.custom_console_handler
        else:
            handler = self.err_handler if record.levelno > logging.INFO and self.multiplex else self.info_handler
        handler.emit(record)
        # ignore broken pipes, this often happens when cleaning up and exiting
        try:
            handler.flush()
        except BrokenPipeError:
            pass

    def flush(self):
        try:
            self.err_handler.flush()
            self.info_handler.flush()
        except BrokenPipeError:
            pass

    def close(self):
        self.err_handler.close()
        self.info_handler.close()

    def setFormatter(self, fmt):
        self.err_handler.setFormatter(fmt)
        self.info_handler.setFormatter(fmt)

class StimelaLogFormatter(logging.Formatter):
    DEBUG_STYLE   = "dim", ""
    WARNING_STYLE = "", "yellow"
    ERROR_STYLE   = "bold", "red"
    CRITICAL_STYLE  = "bold", "red"

    _ansi_escape = re.compile(r'\x1B[@-_][0-?]*[ -/]*[@-~]')

    def __init__(self, boring=False):
        datefmt = "%Y-%m-%d %H:%M:%S"
        if not boring:
            super().__init__("{asctime} {name} [{style}]{levelname}: {message}[/{style}]", datefmt, style="{")
            self._prefix_fmt = logging.Formatter("[{style}]{prefix} {message}[/{style}]", style="{")
        else:
            super().__init__("{asctime} {name} {levelname}: {message}", datefmt, style="{")
            self._prefix_fmt = logging.Formatter("{prefix} {message}", style="{")
        self.boring = boring

    def format(self, record):
        # apply styling
        record = copy.copy(record)
        if not hasattr(record, 'style'):
            if record.levelno <= logging.DEBUG:
                font, color = self.DEBUG_STYLE
            elif record.levelno >= logging.CRITICAL:
                font, color = self.SEVERE_STYLE
            elif record.levelno >= logging.ERROR:
                font, color = self.ERROR_STYLE
            elif record.levelno >= logging.WARNING:
                font, color = self.WARNING_STYLE
            else:
                font = color = ""
            if getattr(record, 'color', None):
                color = record.color.lower()
            if hasattr(record, 'bold') or hasattr(record, 'boldface'):
                if "bold" not in font:
                    font = f"bold {font}"
            setattr(record, 'style', f"{font} {color}".strip() or "normal")
        # in boring mode, make funky messages boring, and strip ANSI codes
        if self.boring:
            record.msg = self._ansi_escape.sub('', defunkify(record.msg))
        # select format based on whether we have prefix or not
        if hasattr(record, 'prefix'):
            if self.boring:
                record.prefix = defunkify(record.prefix)
            return self._prefix_fmt.format(record)
        else:
            return super().format(record)

_logger = None
log_console_handler = log_formatter = log_file_formatter = log_boring_formatter = log_colourful_formatter = None

LOG_DIR = '.'

def is_logger_initialized():
    return _logger is not None


def declare_chapter(title: str, **kw):
    progress_console.rule(title, **kw)


def logger(name="STIMELA", propagate=False, boring=False, loglevel="INFO"):
    """Returns the global Stimela logger (initializing if not already done so, with the given values)"""
    global _logger
    if _logger is None:
        _logger = logging.getLogger(name)
        if type(loglevel) is str:
            loglevel = getattr(logging, loglevel)
        _logger.setLevel(loglevel)
        _logger.propagate = propagate

        global log_console_handler, log_formatter, log_file_formatter, log_boring_formatter, log_colourful_formatter
        global progress_console, progress_bar

        log_boring_formatter = StimelaLogFormatter(boring=True)
        log_colourful_formatter = StimelaLogFormatter(boring=False)

        log_formatter = log_boring_formatter if boring else log_colourful_formatter

        progress_bar, progress_console = task_stats.init_progress_bar()

        log_console_handler = rich.logging.RichHandler(console=progress_console,
                            highlighter=rich.highlighter.NullHighlighter(), markup=True,
                            show_level=False, show_path=False, show_time=False, keywords=[])

        log_console_handler.setFormatter(log_formatter)
        log_console_handler.setLevel(loglevel)

        _logger.addHandler(log_console_handler)
        _logger_console_handlers[_logger.name] = log_console_handler

        import scabha
        scabha.set_logger(_logger)

    return _logger

_logger_file_handlers = {}
_logger_console_handlers = {}


def has_file_logger(log: logging.Logger):
    return log.name in _logger_file_handlers


def disable_file_logger(log: logging.Logger):
    current_logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    if fh is not None:
        fh.close()
        log.removeHandler(fh)
        del _logger_file_handlers[log.name]


class DelayedFileHandler(logging.FileHandler):
    """A version of FileHandler that also handles directory and symlink creation in a delayed way"""
    def __init__(self, logfile, symlink, mode):
        self.symlink, self.logfile = symlink, logfile
        self.is_open = False
        super().__init__(logfile, mode, delay=True)

    def get_logfile_dir(self):
        """Gets name of logfile and ensures the directory exists"""
        if not self.is_open:
            self.is_open = True
            logdir = os.path.dirname(self.logfile)
            if logdir and not os.path.exists(logdir):            
                os.makedirs(logdir)
                if self.symlink:
                    symlink_path = os.path.join(os.path.dirname(logdir.rstrip("/")) or ".", self.symlink)
                    # remove existing symlink
                    if os.path.islink(symlink_path):
                        os.unlink(symlink_path)
                    # Make symlink to logdir. If name exists and is not a symlink, we'll do nothing
                    if not os.path.exists(symlink_path):
                        os.symlink(os.path.basename(logdir), symlink_path)
        return os.path.dirname(self.logfile)

    def emit(self, record):
        self.get_logfile_dir()
        return super().emit(record)

def setup_file_logger(log: logging.Logger, logfile: str, level: Optional[Union[int, str]] = logging.INFO, symlink: Optional[str] = None):
    """Sets up logging to file

    Args:
        log (logging.Logger): Logger object
        logfile (str): logfile. May contain dirname, which will be created as needed.
        level (Optional[Union[int, str]], optional): Logging level, defaults to logging.INFO.
        symlink (Optional[str], optional): if set, and logfile contains a dirname that is created, sets named symlink to point to it
            (This is useful for patterns such as logfile="logs-YYMMDD/logfile.txt", then logs -> logs-YYMMDD)

    Returns:
        [logging.Logger]: logger object
    """
    current_logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    
    # does the logger need a new FileHandler created
    if current_logfile != logfile:
        log.debug(f"will switch to logfile {logfile} (previous was {current_logfile})")
        # remove old FH if so
        if fh is not None:
            fh.close()
            log.removeHandler(fh)

        # create new FH
        fh = DelayedFileHandler(logfile, symlink, 'w')
        fh.setFormatter(log_boring_formatter)
        log.addHandler(fh)

        _logger_file_handlers[log.name] = logfile, fh

        # if logging to console, disable propagation from this sub-logger, and add a console handler
        # This ensures that parent loggers that log to files to not get repeated messages
        if log_console_handler:
            log.propagate = False
            if log.name not in _logger_console_handlers:
                _logger_console_handlers[log.name] = log_console_handler
                log.addHandler(log_console_handler)


    # resolve level
    if level is not None:
        if type(level) is str:
            level = getattr(logging, level, logging.INFO)
        fh.setLevel(level)

    return log


def update_file_logger(log: logging.Logger, logopts: DictConfig, nesting: int = 0, subst: Optional[SubstitutionNS] = None, location=[]):
    """Updates logfiles associated with given logger based on option settings

    Args:
        log (logging.Logger):                          Logger object
        nesting (int):                                 nesting level of this logger
        logopts (Union[StimelaLogConfig, DictConfig]): config settings
        subst (Dict[str, Any]):                        dictionary of substitutions for pathnames in logopts
        location (List[str]):                          location of this logger in the hierarchy  

    Returns:
        [type]: [description]
    """

    if logopts.enable and logopts.nest >= nesting:
        path = os.path.join(logopts.dir or ".", logopts.name + logopts.ext)

        if subst is not None:
            with forgiving_substitutions_from(subst, raise_errors=False) as context: 
                path = context.evaluate(path, location=location + ["log"])
                if context.errors:
                    for err in context.errors:
                        log.error(f"bad substitution in log path: {err}")
                    return None

        # substitute non-filename characters for _
        path = re.sub(r'[^a-zA-Z0-9_./-]', '_', path)

        # setup the logger
        setup_file_logger(log, path, level=logopts.level, symlink=logopts.symlink)
    else:
        disable_file_logger(log)


def get_logfile_dir(log: logging.Logger):
    """Returns filename associated with the logger, or None if not logging to file"""
    logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    if logfile is None:
        return None
    return fh.get_logfile_dir()


def log_exception(*errors, severity="error", log=None):
    """Logs one or more error messages or exceptions (unless they are marked as already logged), and 
    pretty-prints them to the console  as appropriate.
    """
    def exc_message(e):
        if isinstance(e, ScabhaBaseException):
            return escape(e.message)
        elif type(e) is str:
            return escape(e)
        else:
            return escape(f"{type(e).__name__}: {e}")

    if severity == "error":
        colour = "bold red"
        message_dispatch = (log or logger()).error
    else:
        colour = "yellow"
        message_dispatch = (log or logger()).warning

    trees = []
    do_log = False
    messages = []

    def add_dict(dd, tree):
        for field, value in dd.items():
            if isinstance(value, (dict, OrderedDict, DictConfig)):
                subtree = tree.add(escape(f"{field}:"))
                add_dict(value, subtree)
            else:
                tree.add(escape(f"{field}: {value}"))

    def add_nested(excs, tree):
        for exc in excs:
            if isinstance(exc, Exception):
                subtree = tree.add(f"{exc_message(exc)}")
                # tbtree = subtree.add("Traceback:")
                # for line in traceback.format_exception(exc):
                #     tbtree.add(line)
                if isinstance(exc, ScabhaBaseException) and exc.nested:
                    add_nested(exc.nested, subtree)
            elif type(exc) is TracebackType:
                subtree = tree.add(f"[dim]Traceback:[/dim]")
                for line in traceback.format_tb(exc):
                    subtree.add(f"[dim]{escape(line.rstrip())}[/dim]")
            elif type(exc) is FormattedTraceback:
                subtree = tree.add(f"[dim]Traceback:[/dim]")
                for line in exc.lines:
                    subtree.add(f"[dim]{escape(line)}[/dim]")
            elif isinstance(exc, (dict, OrderedDict, DictConfig)):
                add_dict(exc, tree)
            else:
                tree.add(str(exc))

    has_nesting = False

    for exc in errors:
        if isinstance(exc, ScabhaBaseException):
            messages.append(exc.message)
            if not exc.logged:
                do_log = exc.logged = True
            tree = Tree(f"[{colour}]:warning: {exc_message(exc)}[/{colour}]", guide_style="dim")
            trees.append(tree)
            if exc.nested:
                add_nested(exc.nested, tree)
                has_nesting = True
        else:
            tree = Tree(f"[{colour}]:warning: {exc_message(exc)}[/{colour}]", guide_style="dim")
            trees.append(tree)
            do_log = True
            messages.append(str(exc))

    if do_log:
        message_dispatch(": ".join(messages))

    printfunc = task_stats.progress_bar.console.print if task_stats.progress_bar is not None else rich_print

    if has_nesting:
        declare_chapter("detailed error report follows", style="red")
        for tree in trees:
            printfunc(Padding(tree, pad=(0,0,0,8)))





import subprocess
import os
import platform
import getpass
from io import StringIO
from stimela import utils
import json
import stimela
import time
import datetime
import subprocess
import yaml
from typing import Dict, Union
from stimela.config import StimelaImage

class DockerError(Exception):
    pass

from stimela.backends import StimelaImageBuildInfo, StimelaImageInfo

_available_images: Union[Dict[str,Dict[str, StimelaImageInfo]], None] = None

def available_images():
    """Scans system for available stimela images and returns dicitonary of StimelaImageInfo objects.

    Stimela docker images are identified by a stimela.image.name label.

    Returns
    -------
    dict
        dictionary: {image_name: {version: image_info}}  
    """
    from stimela.main import log
    global _available_images
    if _available_images is None:

        # get list of image IDs which have the right label
        proc = subprocess.run(["docker", "images",
                            "--filter", "label=stimela.image.name", 
                            "--format", "{{.ID}}"], stdout=subprocess.PIPE)
        _available_images = {}
        iids = proc.stdout.split()

        # inspect details of matching IDs
        if iids:
            proc = subprocess.run(["docker", "inspect"] + iids, 
                                stdout=subprocess.PIPE)
            # parse output
            inspect_data = yaml.safe_load(proc.stdout)
            for num, image_data in enumerate(inspect_data):
                iid = image_data.get('Id')
                repotags = image_data.get('RepoTags')
                if iid is None:
                    log.warning(f"failed to parse 'docker inspect' output element {num} {repotags}, skipping")
                    continue
                try:
                    labels = image_data['ContainerConfig']['Labels']
                    name = labels['stimela.image.name']
                    version = labels['stimela.image.version']
                    build = {}
                    for key in 'stimela_version', 'user', 'host', 'date':
                        build[key] = labels[f'stimela.build.{key}']
                except KeyError as keyerr:
                    log.warning(f"failed to parse 'docker inspect' output element {num} {repotags}: missing key {keyerr}, skipping")
                    continue

                _available_images.setdefault(name, {})[version] = StimelaImageInfo(name=name, version=version, iid=iid, 
                                                                                   full_name=repotags[0], build=StimelaImageBuildInfo(**build))
        
    return _available_images


def _get_full_name(image: StimelaImage, version:str):
    """Returns full image name (e.g. "quay.io/stimela/v2-NAME:VERSION")

    Parameters
    ----------
    image : StimelaImage
        image object
    version : str
        version
    """
    from stimela.main import CONFIG
    if CONFIG.opts.registry:
        basename = f"{CONFIG.opts.registry}/{CONFIG.opts.basename}"
    else:
        basename = CONFIG.opts.basename
    return f"{basename}{image.name}:{version}"


def build(image: StimelaImage, version: str):
    """Builds given image + version

    Parameters
    ----------
    image : StimelaImage
        image object
    version : str
        version to be built, must be present in image.images
    """
    from stimela.main import log

    fullname = _get_full_name(image, version)

    build_info = image.images[version]
    cwd = os.path.dirname(image.path)
    dockerfile = os.path.join(cwd, build_info.dockerfile)
    log.info(f"building {fullname} using {dockerfile}")

    subprocess.run(["docker", "build", "-t", fullname, "-f", dockerfile,
                    "--label", f"stimela.image.name={image.name}", 
                    "--label", f"stimela.image.version={version}", 
                    "--label", f"stimela.build.stimela_version={stimela.__version__}", 
                    "--label", f"stimela.build.user={getpass.getuser()}", 
                    "--label", f"stimela.build.host={platform.node()}", 
                    "--label", f"stimela.build.date={datetime.datetime.now().ctime()}", 
                    cwd], check=True)

    # reset this to force a rescan in available_images()
    global _available_images
    _available_images = None


def push(image: StimelaImage, version: str):
    """Pushes given image + version to registry

    Parameters
    ----------
    image : StimelaImage
        image object
    version : str
        version to be pushed
    """
    from stimela.main import log

    fullname = _get_full_name(image, version)
    log.info(f"pushing {fullname}")

    subprocess.run(["docker", "push", fullname], check=True)
        


def pull(image, tag=None, force=False):
    """ pull a docker image """
    if tag:
        image = ":".join([image, tag])

    utils.xrun("docker", ["pull", image])


def seconds_hms(seconds):
    return str(datetime.timedelta(seconds=seconds))


class Container(object):
    def __init__(self, image, name,
                 volumes=None, environs=None,
                 label="", logger=None,
                 time_out=-1,
                 workdir=None,
                 log_container=None,
                 cabname=None,
                 runscript=None):
        """
        Python wrapper to docker engine tools for managing containers.
        """

        self.image = image
        self.name = name
        self.cabnane = cabname
        self.label = label
        self.volumes = volumes or []
        self.environs = environs or []
        self.logger = logger
        self.status = None
        self.WORKDIR = workdir
        self.RUNSCRIPT = runscript
        self.PID = os.getpid()
        self.uptime = "00:00:00"
        self.time_out = time_out
        self.cont_logger = utils.logger.StimelaLogger(
            log_container or stimela.LOG_FILE, jtype="docker")

    def add_volume(self, host, container, perm="rw", noverify=False):

        if os.path.exists(host) or noverify:
            if self.logger:
                self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
                    host, self.name, container))
            host = os.path.abspath(host)
        else:
            raise IOError(
                "Directory {0} cannot be mounted on container: File doesn't exist".format(host))

        self.volumes.append(":".join([host, container, perm]))

    def add_environ(self, key, value):
        if self.logger:
            self.logger.debug("Adding environ varaible [{0}={1}] in container {2}".format(
                key, value, self.name))
        self.environs.append("=".join([key, value]))

    def create(self, *args):

        if self.volumes:
            volumes = " -v " + " -v ".join(self.volumes)
        else:
            volumes = ""
        if self.environs:
            environs = environs = " -e "+" -e ".join(self.environs)
        else:
            environs = ""

        self._print(
            "Instantiating container [{}]. The container ID is printed below.".format(self.name))
        utils.xrun("docker create", list(args) + [volumes, environs, "--rm",
                                                  "-w %s" % (self.WORKDIR),
                                                  "--name", self.name,
                                                  self.image,
                                                  self.RUNSCRIPT or ""], log=self.logger)

        self.status = "created"

    def info(self):

        output = subprocess.check_output(
            "docker inspect {}".format(self.name), shell=True).decode()
        output_file = StringIO(output[3:-3])
        jdict = json.load(output_file)
        output_file.close()

        return jdict

    def get_log(self):
        stdout = open(self.logfile, 'w')
        exit_status = subprocess.call("docker logs {0}".format(self.name),
                                      stdout=stdout, stderr=stdout, shell=True)
        if exit_status != 0:
            self.logger.warn(
                'Could not log container: {}. Something went wrong durring execution'.format(self.name))
            output = 'Task was not started.'
            stdout.write(output)
        else:
            output = stdout.read()

        stdout.close()
        return output

    def start(self, output_wrangler=None):
        running = True
        tstart = time.time()
        self.status = "running"

        self.cont_logger.log_container(self.name)
        self.cont_logger.write()
        self._print("Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
            self.name, self.time_out))
        utils.xrun("docker", ["start", "-a", self.name],
                       timeout=self.time_out,
                       logfile=self.logfile,
                       log=self.logger, output_wrangler=output_wrangler,
                       kill_callback=lambda: utils.xrun("docker", ["kill", self.name]))
        uptime = seconds_hms(time.time() - tstart)
        self.uptime = uptime
        self._print(
            "Container [{0}] has executed successfully".format(self.name))

        self._print("Runtime was {0}.".format(uptime))

        self.status = "exited"

    def stop(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status in ["running", "paused"]:
            try:
                utils.xrun("docker stop", [self.name])
            except KeyboardInterrupt("Received terminate signal. Will stop and remove container first"):
                killed = True
        self.status = 'exited'

        self._print("Container {} has been stopped.".format(self.name))
        if killed:
            self.remove()
            raise KeyboardInterrupt

    def image_exists(self):
        """
            Check if image exists 
        """
        image_ids = subprocess.check_output(f"docker images -q {self.image}".split())
        if image_ids:
            return True
        else:
            return False

    def remove(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status == "exited":
            try:
                utils.xrun("docker rm", [self.name])
            except KeyboardInterrupt:
                killed = True
            if killed:
                raise KeyboardInterrupt

        else:
            raise DockerError(
                "Container [{}] has not been stopped, cannot remove".format(self.name))

        self.cont_logger.remove('containers', self.name)
        self.cont_logger.write()

    def _print(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)



import dataclasses
import re, importlib, sys
from collections import OrderedDict
from enum import Enum, IntEnum
from dataclasses import dataclass
from omegaconf import MISSING, ListConfig, DictConfig, OmegaConf

import rich.box
import rich.markup
from rich.table import Table
from rich.markdown import Markdown

from .exceptions import ParameterValidationError, DefinitionError, SchemaError, AssignmentError
from .validate import validate_parameters, Unresolved
from .substitutions import SubstitutionNS
from .basetypes import EmptyDictDefault, EmptyListDefault, UNSET

# need * imports from both to make eval(self.dtype, globals()) work
from typing import *
from .basetypes import *

## almost supported by omegaconf, see https://github.com/omry/omegaconf/issues/144, for now just use Any
ListOrString = Any

Conditional = Optional[str]

# this marks unset defaults: it's turned into an UNSET by the __post_init__
# method (UNSET can't be used directly in an OmegeConf structured schema, hence
# the need for this stopgap)
_UNSET_DEFAULT = "<UNSET DEFAULT VALUE>"

@dataclass
class ParameterPolicies(object):
    """This class describes policies that determine how a Parameter is turned into
    cab arguments. Most policies refer to how command-line arguments are formed up,
    although some also apply to Python callable cabs.
    """
    # if true, parameter is passed as key=value, not command line option
    key_value: Optional[bool] = None
    # if true, value is passed as a positional argument, not an option
    positional: Optional[bool] = None
    # if true, value is head-positional, i.e. passed *before* any options
    positional_head: Optional[bool] = None
    # for list-type values, use this as a separator to paste them together into one argument. Otherwise:
    #  * use "list" to pass list-type values as multiple arguments (--option X Y)
    #  * use "[]" to pass list-type values as a list  (--option [X,Y])
    #  * use "repeat" to repeat the option (--option X --option Y)
    repeat: Optional[str] = None
    # prefix for non-positional arguments
    prefix: Optional[str] = None

    # skip this parameter
    skip: Optional[bool] = None
    # if True, implicit parameters will be skipped 
    skip_implicits: Optional[bool] = None

    # if set, {}-substitutions on this paramater will not be done
    disable_substitutions: Optional[bool] = None

    # how to pass boolean True values. None = pass option name alone, else pass option name + given value
    explicit_true: Optional[str] = None
    # how to pass boolean False values. None = skip option, else pass option name + given value
    explicit_false: Optional[str] = None

    # if set, a string-type value will be split into a list of arguments using this separator
    split: Optional[str] = None

    # dict of character replacements for mapping parameter name to command line
    replace: Optional[Dict[str, str]] = None

    # Value formatting policies.
    # If set, specifies {}-type format strings used to convert the value(s) to string(s).
    # For a non-list value:
    #   * if 'format_list_scalar' is set, formats the value into a list of strings as fmt[i].format(value, **dict)
    #     example:  ["{0}", "{0}"] will simply repeat the value twice
    #   * if 'format' is set, value is formatted as format.format(value, **dict)
    # For a list-type value:
    #   * if 'format_list' is set, each element #i formatted separately as fmt[i].format(*value, **dict)
    #     example:  ["{0}", "{2}"] will output elements 0 and 2, and skip element 1
    #   * if 'format' is set, each element #i is formatted as format.format(value[i], **dict)
    # **dict contains all parameters passed to a cab, so these can be used in the formatting
    format: Optional[str] = None
    format_list: Optional[List[str]] = None
    format_list_scalar: Optional[List[str]] = None

    # for Python callable cabs: if set, then missing parameters are passed as None values
    # if not set, missing parameters are not passed at all
    pass_missing_as_none: Optional[bool] = None


# This is used to classify parameters, for cosmetic and help purposes.
# Usually set automatically based on whether a parameter is required, whether a default is provided, etc.
ParameterCategory = IntEnum("ParameterCategory",
                            dict(Required=0, Optional=1, Implicit=2, Obscure=3, Hidden=4),
                            module=__name__)

@dataclass
class Parameter(object):
    """Parameter (of cab or recipe)"""
    info: str = ""
    # for input parameters, this flag indicates a read-write (aka input-output aka mixed-mode) parameter e.g. an MS
    writable: bool = False
    # data type
    dtype: str = "str"
    # specifies that the value is implicitly set inside the step (i.e. not a free parameter). Typically used with filenames
    implicit: Any = None
    # optonal list of arbitrary tags, used to group parameters
    tags: List[str] = EmptyListDefault()

    # If True, parameter is required. None/False, not required.
    # For outputs, required=False means invalid output will not be treated as an error.
    # For aliases, False at recipe level will override the target setting, while the default of None won't.
    required: Optional[bool] = None

    # restrict value choices, i.e. making for an option-type parameter
    choices:  Optional[List[Any]] = ()

    # for List or Dict-type parameters, restict values of list elements or dict entries to a list of choices
    element_choices: Optional[List[Any]] = None

    # default value
    default: Any = _UNSET_DEFAULT

    # list of aliases for this parameter (i.e. references to other parameters whose schemas/values this parameter shares)
    aliases: Optional[List[str]] = ()

    # if true, treat parameter as a path, and ensure that the parent directories it refers to exist
    mkdir: bool = False

    # for file and dir-type parameters: if True, the file(s)/dir(s) must exist. If False, they can be missing.
    # if None, then the default logic applies: inputs must exist, and outputs don't
    must_exist: Optional[bool] = None

    # For output File-type parameters. If True, and the output exists, remove before running
    remove_if_exists: bool = False

    # if command-line option for underlying binary has a different name, specify it here
    nom_de_guerre: Optional[str] = None

    # policies object, specifying a non-default way to handle this parameter
    policies: ParameterPolicies = ParameterPolicies()

    # Parameter category, purely cosmetic, used for generating help and debug messages.
    # Assigned automatically if None, but a schema may explicitly mark parameters as e.g.
    # "obscure" or "hidden"
    category: Optional[ParameterCategory] = None

    # metavar corresponding to this parameter. Used when constructing command-line interfaces
    metavar: Optional[str] = None

    # abbreviated option name for this parameter.  Used when constructing command-line interfaces
    abbreviation: Optional[str] = None

    # arbitrary metadata associated with parameter
    metadata: Dict[str, Any] = EmptyDictDefault()

    def __post_init__(self):
        def natify(value):
            # convert OmegaConf lists and dicts to native types
            if type(value) in (list, ListConfig):
                return [natify(x) for x in value]
            elif type(value) in (dict, OrderedDict, DictConfig):
                return OrderedDict([(name, natify(value)) for name, value in value.items()])
            elif value is _UNSET_DEFAULT:
                return UNSET
            return value
        self.default = natify(self.default)
        self.choices = natify(self.choices)

        # converts string dtype into proper type object
        # yes I know eval() is naughty but this is the best we can do for now
        # see e.g. https://stackoverflow.com/questions/67500755/python-convert-type-hint-string-representation-from-docstring-to-actual-type-t
        # The alternative is a non-standard API call i.e. typing._eval_type()
        self._dtype = eval(self.dtype, globals())

        self._is_input = True

    def get_category(self):
        """Returns category of parameter, auto-setting it if not already preset"""
        if self.category is None:
            if self.required:
                self.category = ParameterCategory.Required
            elif self.implicit is not None:
                self.category = ParameterCategory.Implicit
            else:
                self.category = ParameterCategory.Optional
        return self.category

    @property 
    def is_input(self):
        return self._is_input

    @property 
    def is_output(self):
        return not self._is_input

    _filename_types = (File, MS, Directory, "File", "MS", "Directory")

    @property
    def is_file_type(self):
        """True if parameter is a file or directory type"""
        return self.dtype in self._filename_types

    @property
    def is_named_output(self):
        """True if parameter is a named file or directory output"""
        return self.is_output and self.is_file_type and not self.implicit

ParameterSchema = OmegaConf.structured(Parameter)

ParameterFields = set(f.name for f in dataclasses.fields(Parameter))

@dataclass
class Cargo(object):
    name: Optional[str] = None                    # cab name (if None, use image or command name)
    fqname: Optional[str] = None                  # fully-qualified name (recipe_name.step_label.etc.etc.)

    info: Optional[str] = None                    # description

    # schemas are postentially nested (dicts of dicts), which omegaconf doesn't quite recognize,
    # (or in my ignorance I can't specify it -- in any case Union support is weak), so do a dict to Any
    # "Leaf" elements of the nested dict must be Parameters
    inputs: Dict[str, Any]   = EmptyDictDefault()
    outputs: Dict[str, Any]  = EmptyDictDefault()
    defaults: Dict[str, Any] = EmptyDictDefault()

    backend: Optional[str] = None                 # backend, if not default

    dynamic_schema: Optional[str] = None          # function to call to augment inputs/outputs dynamically

    @staticmethod
    def flatten_schemas(io_dest, io, label, prefix=""):
        for name, value in io.items():
            if name == "subsection":
                continue
            name = f"{prefix}{name}"
            if not isinstance(value, Parameter):
                if not isinstance(value, (DictConfig, dict)):
                    raise SchemaError(f"{label}.{name} is not a valid schema")
                # try to treat as Parameter based on field names
                if not (set(value.keys()) - ParameterFields):
                    try:
                        value = OmegaConf.unsafe_merge(ParameterSchema.copy(), value)
                        io_dest[name] = Parameter(**value)
                    except Exception as exc0:
                        raise SchemaError(f"{label}.{name} is not a valid parameter definition", exc0)
                else:
                    Cargo.flatten_schemas(io_dest, value, label=label, prefix=f"{name}.")
        return io_dest

    def flatten_param_dict(self, output_params, input_params, prefix=""):
        for name, value in input_params.items():
            name = f"{prefix}{name}"
            if isinstance(value, (dict, DictConfig)):
                # if prefix.name. is present in schemas, treat as nested mapping
                if any(k.startswith(f"{name}.") for k in self.inputs_outputs):
                    self.flatten_param_dict(output_params, value, prefix=f"{name}.")
                    continue
            output_params[name] = value
        return output_params

    def __post_init__(self):
        self.fqname = self.fqname or self.name
        # flatten inputs/outputs into a single dict (with entries like sub.foo.bar)
        self.inputs = Cargo.flatten_schemas(OrderedDict(), self.inputs, "inputs")
        self.outputs = Cargo.flatten_schemas(OrderedDict(), self.outputs, "outputs")
        for schema in self.outputs.values():
            schema._is_input = False
        for name in self.inputs.keys():
            if name in self.outputs:
                raise DefinitionError(f"parameter '{name}' appears in both inputs and outputs")
        self._inputs_outputs = None
        self._implicit_params = set()   # marks implicitly set values
        # flatten defaults and aliases
        self.defaults = self.flatten_param_dict(OrderedDict(), self.defaults)
        # pausterized name
        self.name_ = re.sub(r'\W', '_', self.name or "")  # pausterized name
        # config and logger objects
        self.config = self.log = self.logopts = None
        # resolve callable for dynamic schemas
        self._dyn_schema = None
        if self.dynamic_schema is not None:
            if '.' not in self.dynamic_schema:
                raise DefinitionError(f"{self.dynamic_schema}: module_name.function_name expected")
            modulename, funcname = self.dynamic_schema.rsplit(".", 1)
            try:
                mod = importlib.import_module(modulename)
            except ImportError as exc:
                raise DefinitionError(f"can't import {modulename}: {exc}")
            self._dyn_schema = getattr(mod, funcname, None)
            if not callable(self._dyn_schema):
                raise DefinitionError(f"{modulename}.{funcname} is not a valid callable")

    @property
    def inputs_outputs(self):
        if self._inputs_outputs is None:
            self._inputs_outputs = self.inputs.copy()
            self._inputs_outputs.update(**self.outputs)
        return self._inputs_outputs

    @property
    def finalized(self):
        return self.config is not None

    def unresolved_params(self, params):
        """Returns list of unresolved parameters"""
        return [name for name, value in params.items() if isinstance(value, Unresolved)]


    def finalize(self, config=None, log=None, fqname=None, nesting=0):
        if not self.finalized:
            if fqname is not None:
                self.fqname = fqname
            self.config = config
            self.nesting = nesting
            self.log = log
            self.logopts = config.opts.log.copy()

    def apply_dynamic_schemas(self, params):
        # update schemas, if dynamic schema is enabled
        if self._dyn_schema:
            self._inputs_outputs = None
            try:
                self.inputs, self.outputs = self._dyn_schema(params, self.inputs, self.outputs)
            except Exception as exc:
                raise SchemaError(f"error evaluating dynamic schema", exc) # [exc, sys.exc_info()[2]])
            for io in self.inputs, self.outputs:
                for name, schema in list(io.items()):
                    if isinstance(schema, DictConfig):
                        try:
                            schema = OmegaConf.unsafe_merge(ParameterSchema.copy(), schema)
                        except Exception  as exc:
                            raise SchemaError(f"error in dynamic schema for parameter 'name'", exc)
                        io[name] = Parameter(**schema)

    def prevalidate(self, params: Optional[Dict[str, Any]], subst: Optional[SubstitutionNS]=None, root=False):
        """Does pre-validation.
        No parameter substitution is done, but will check for missing params and such.
        A dynamic schema, if defined, is applied at this point."""
        self.finalize()
        # add implicits, if resolved
        for name, schema in self.inputs_outputs.items():
            if schema.implicit is not None and type(schema.implicit) is not Unresolved:
                if name in params and name not in self._implicit_params:
                    raise ParameterValidationError(f"implicit parameter {name} was supplied explicitly")
                if name in self.defaults:
                   raise SchemaError(f"implicit parameter {name} also has a default value")
                params[name] = schema.implicit
                self._implicit_params.add(name)
        # assign unset categories
        for name, schema in self.inputs_outputs.items():
            schema.get_category()

        params = validate_parameters(params, self.inputs_outputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                          check_unknowns=True, check_required=False, check_exist=False,
                                          create_dirs=False, ignore_subst_errors=True)

        return params

    def validate_inputs(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, loosely=False):
        """Validates inputs.
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        This is used when skipping a step.
        """
        assert(self.finalized)

        # check inputs
        params1 = validate_parameters(params, self.inputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                                check_unknowns=False, check_required=not loosely, check_exist=not loosely,
                                                create_dirs=not loosely)
        # check outputs
        params1.update(**validate_parameters(params, self.outputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                                check_unknowns=False, check_required=False, check_exist=False,
                                                create_dirs=not loosely))
        return params1

    def validate_outputs(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, loosely=False):
        """Validates outputs. Parameter substitution is done.
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        """
        assert(self.finalized)
        # update implicits that weren't marked as unresolved
        for name in self._implicit_params:
            impl = self.inputs_outputs[name].implicit
            if type(impl) is not Unresolved:
                params[name] = self.inputs_outputs[name].implicit
        params.update(**validate_parameters(params, self.outputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                                check_unknowns=False, check_required=not loosely, check_exist=not loosely))
        return params

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        """Generates help into a rich.tree.Tree object"""
        if self.info:
            tree.add("Description:").add(Markdown(self.info))
        # adds tables for inputs and outputs
        for io, title in (self.inputs, "inputs"), (self.outputs, "outputs"):
            for cat in ParameterCategory:
                schemas = [(name, schema) for name, schema in io.items() if schema.get_category() == cat]
                if not schemas:
                    continue
                if cat > max_category:
                    subtree = tree.add(f"[dim]{cat.name} {title}: omitting {len(schemas)}[/dim]")
                    continue
                subtree = tree.add(f"{cat.name} {title}:")
                table = Table.grid("", "", "", padding=(0,2)) # , show_header=False, show_lines=False, box=rich.box.SIMPLE)
                subtree.add(table)
                for name, schema in schemas:
                    attrs = []
                    default = self.defaults.get(name, schema.default)
                    if schema.implicit:
                        attrs.append(f"implicit: {schema.implicit}")
                    if default is not UNSET and not isinstance(default, Unresolved):
                        attrs.append(f"default: {default}")
                    if schema.choices:
                        attrs.append(f"choices: {', '.join(schema.choices)}")
                    info = []
                    schema.info and info.append(rich.markup.escape(schema.info))
                    attrs and info.append(f"[dim]\[{rich.markup.escape(', '.join(attrs))}][/dim]")
                    table.add_row(f"[bold]{name}[/bold]",
                                  f"[dim]{rich.markup.escape(str(schema.dtype))}[/dim]",
                                  " ".join(info))

    def assign_value(self, key: str, value: Any, override: bool = False):
        """assigns a parameter value to the cargo. 
        Recipe will override this to handle nested assignments. Cabs can't be assigned to
        (it will be handled by the wraping step)
        """
        raise AssignmentError(f"{self.name}: invalid assignment {key}={value}")




import re, os.path, json, zlib, codecs, base64
from typing import Optional, Any, Union, Dict
from dataclasses import dataclass

import stimela
from scabha.exceptions import SubstitutionError
from stimela.exceptions import CabValidationError
from stimela.kitchen.cab import Cab
from scabha.cab_utils import CAB_OUTPUT_PREFIX
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour, _BaseFlavour


def form_python_function_call(function: str, cab: Cab, params: Dict[str, Any]):
    """
    Helper. Converts a function name and a list of parameters into a string 
    representation of a Python function call that can be parsed by the interpreter.
    Uses the cab schema info and policies to decide which parametets to include.

    Args:
        function (str) function name
        cab (Cab): cab definition
        params (Dict[str, Any]): dict of parameters, e.g. {a: 1, b: 'foo'}

    Returns:
        str: function invocation, e.g. "func(a=1, b='foo')"
    """
    arguments = []
    for key, value in cab.filter_input_params(params).items():
        arguments.append(f"{key}={repr(value)}")
    return f"{function}({', '.join(arguments)})"


def get_python_interpreter_args(cab: Cab, subst: Dict[str, Any]):
    """
    Helper. Given a cab definition, forms up appropriate argument list to
    invoke the interpreter. Invokes a virtual environment as appropriate.

    Args:
        cab (Cab):              cab definition 
        subst (Dict[str, Any]): substitution namespace

    Raises:
        CabValidationError: on errors

    Returns:
        List[str]: [command, arguments, ...] needed to invoke the interpreter
    """    
    # get virtual env, if specified
    with substitutions_from(subst, raise_errors=True) as context:
        venv = context.evaluate(cab.virtual_env, location=["virtual_env"])

    if venv:
        venv = os.path.expanduser(venv)
        interpreter = f"{venv}/bin/python"
        if not os.path.isfile(interpreter):
            raise CabValidationError(f"virtual environment {venv} doesn't exist")
    else:
        interpreter = "python"

    return [interpreter, "-u"]


@dataclass
class PythonCallableFlavour(_CallableFlavour):
    """
    Represents a cab flavour that is a Python callable. Cab command field is
    expected to be in the form of [package.]module.function
    """
    kind: str = "python"
    # don't log full command by default, as that's full of code
    log_full_command: bool = False

    def finalize(self, cab: Cab):
        super().finalize(cab)
        # form up outputs handler
        if self.output is not None or self.output_dict:
            self._yield_output = f"print(f'{CAB_OUTPUT_PREFIX}{{json.dumps(_result)}}')"
            pattern = re.compile(f"{CAB_OUTPUT_PREFIX}(.*)")
            if self.output_dict:
                wrangler = wranglers.ParseJSONOutputDict(pattern, "PARSE_JSON")
            else:
                wrangler = wranglers.ParseOutput(pattern, "PARSE_OUTPUT", self.output, '1', "json")
            wrangs = [wrangler]
            if not stimela.VERBOSE:
                wrangs.append(wranglers.Suppress(pattern, "SUPPRESS"))
            cab._wranglers.append((pattern, wrangs))
        else:
            self._yield_output = ""

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any]):
        # substitute command and split into module/function
        with substitutions_from(subst, raise_errors=True) as context:
            try:
                command = context.evaluate(cab.command, location=["command"])
            except Exception as exc:
                raise SubstitutionError(f"error substituting Python callable '{cab.command}'", exc)

        if '.' in command:
            py_module, py_function = cab.command.rsplit('.', 1)
        else:
            raise CabValidationError(f"cab {cab.name}: python flavour requires a command of the form module.function")
        self.command_name = py_function

        # convert inputs into a JSON string
        pass_params = cab.filter_input_params(params)
        params_string = base64.b64encode(
                            zlib.compress(json.dumps(pass_params).encode('ascii'), 2)
                        ).decode('ascii')

        # form up command string
        code = f"""
import sys, json, zlib, base64
_inputs = json.loads(zlib.decompress(
                        base64.b64decode(sys.argv[1].encode("ascii"))
                    ).decode("ascii"))
sys.path.append('.')
from {py_module} import {py_function}
try:
    from click import Command
except ImportError:
    Command = None
if Command is not None and isinstance({py_function}, Command):
    print("invoking callable {command}() (as click command) using external interpreter")
    {py_function} = {py_function}.callback
else:
    print("invoking callable {command}() using external interpreter")
_result = {py_function}(**_inputs)
{self._yield_output}
        """

        args = get_python_interpreter_args(cab, subst)
        args += ["-c", code, params_string]
        return args


@dataclass
class PythonCodeFlavour(_BaseFlavour):
    """
    Represents a cab flavour that is inlined Python code. Cab command field is
    Python code. Parameters can be passed in as local variables, or a dict
    """
    kind: str = "python-code"
    # if set to a string, inputs will be passed in as a dict assigned to a variable of that name
    input_dict: Optional[str] = None
    # if True, inputs will be passed in as named variables
    input_vars: bool = True
    # if True, outputs will be collected from named variables
    output_vars: bool = True
    # if True, command will have {}-substitutions done on it
    subst: bool = False
    # don't log full command by default, as that's full of code
    log_full_command: bool = False

    def finalize(self, cab: Cab):
        super().finalize(cab)
        pattern = re.compile(f"{CAB_OUTPUT_PREFIX}(.*)")
        wrangs = [wranglers.ParseJSONOutputDict(pattern, "PARSE_JSON")]
        if not stimela.VERBOSE:
            wrangs.append(wranglers.Suppress(pattern, "SUPPRESS"))
        cab._wranglers.append((pattern, wrangs))
        self.command_name = "[python]"

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any]):
        # do substitutions on command, if necessary
        if self.subst:
            with substitutions_from(subst, raise_errors=True) as context:
                try:
                    command = context.evaluate(cab.command, location=["command"])
                except Exception as exc:
                    raise SubstitutionError("error substituting inline Python code", exc)
        else:
            command = cab.command

        # only pass inputs and named outputs
        pass_params = cab.filter_input_params(params)

        # form up code to parse params from JSON string that will be given as sys.argv[1]
        params_arg = json.dumps(pass_params)
        pre_command = "import json\n"
        inp_dict = self.input_dict or "_params"
        pre_command = f"""import sys, json
{inp_dict} = json.loads(sys.argv[1])
"""
        if self.input_vars:
            for name in pass_params:
                var_name = name.replace("-", "_").replace(".", "__")
                pre_command += f"""{var_name} = {inp_dict}["{name}"]\n"""

        # form up code to print outputs in JSON
        post_command = ""
        pass_outputs = [name for name, schema in cab.outputs.items()
                        if not schema.is_named_output and not schema.implicit]
        if pass_outputs:
            post_command += "from scabha.cab_utils import yield_output\n"
            if self.output_vars:
                for name in pass_outputs:
                    var_name = name.replace("-", "_").replace(".", "__")
                    post_command += f"yield_output(**{{'{name}': {var_name}}})\n"                

        # form up interpreter invocation
        args = get_python_interpreter_args(cab, subst)
        args += ["-c", pre_command + command + post_command, params_arg]
        return args



from typing import Dict, Optional, Any
from stimela import config

from .step import Step
from .batch import Batch
from .cab import Cab

def run_cab(step: Step, params: Dict[str, Any], 
            backend: 'config.Backend', 
            subst: Optional[Dict[str, Any]] = None, 
            batch: Batch=None) -> Cab.RuntimeStatus:
    log = step.log
    cab = step.cargo
        
    ## NOTE(Sphe)
    # Not sure is this the best implementation
    # But the runtime environment has to default to config.run unless explicitly given in the step
    # But more importantly it has be updated by the Cab environment before running the cab
    
    step.runtime = step.runtime or subst.config.run.copy()
    cab.update_environment(subst=subst)
    step.runtime.env.update(**cab.management.environment)

    backend = __import__(f"stimela.backends.{backend.name}", 
                         fromlist=[backend.name])
    
    return backend.run(cab, runtime=step.runtime, params=params, log=log, subst=subst, batch=batch, fqname=step.fqname)



import re
import os.path
import json
from typing import Optional, Any, Union, Dict
from dataclasses import dataclass

import stimela
from scabha.exceptions import SubstitutionError
from stimela.exceptions import CabValidationError
from stimela.kitchen.cab import Cab
from scabha.cab_utils import CAB_OUTPUT_PREFIX
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour, _BaseFlavour
from .python_flavours import form_python_function_call


def get_python_interpreter_args(cab: Cab, subst: Dict[str, Any]):
    # get virtual env, if specified
    with substitutions_from(subst, raise_errors=True) as context:
        venv = context.evaluate(cab.virtual_env, location=["virtual_env"])

    if venv:
        venv = os.path.expanduser(venv)
        interpreter = f"{venv}/bin/python"
        if not os.path.isfile(interpreter):
            raise CabValidationError(f"virtual environment {venv} doesn't exist")
    else:
        interpreter = "python"

    return [interpreter]


@dataclass
class CasaTaskFlavour(_CallableFlavour):
    kind: str = "casa-task"
    casa: Optional[str] = None
    casa_opts: Optional[str] = None

    def finalize(self, cab: Cab):
        super().finalize(cab)

        err_patt = re.compile("(?P<content>(\tSEVERE\t|ABORTING|\*\*\* Error \*\*\*)(.*))$")
        cab._wranglers.append((
            err_patt, [
                wranglers.DeclareError(err_patt, "ERROR", message="CASA error: {content}" )
            ]
        ))

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any]):
        with substitutions_from(subst, raise_errors=True) as context:
            try:
                command = context.evaluate(cab.command, location=["command"])
            except Exception as exc:
                raise SubstitutionError(f"error substituting casa task '{cab.command}'", exc)
            if self.casa:
                try:
                    casa = context.evaluate(self.casa, location=["casa"])
                except Exception as exc:
                    raise SubstitutionError(f"error substituting casa path '{casa}'", exc)
            else:
                casa = stimela.CONFIG.opts.runtime.get('casa', "casa")
            if self.casa_opts:
                try:
                    casa_opts = context.evaluate(self.casa_opts, location=["casa_opts"])
                except Exception as exc:
                    raise SubstitutionError(f"error substituting casa options '{casa_opts}'", exc)
            else:
                casa_opts = stimela.CONFIG.opts.runtime.get('casa_opts', "--log2term --nologger --nologfile")

        self.command_name = command
        # convert inputs into a JSON string
        pass_params = cab.filter_input_params(params)
        params_string = json.dumps(pass_params)

        code = f"""
import sys, json;
kw=json.loads(sys.argv[-1]);
def stringify(x):
    return str(x) if type(x) is unicode else ([stringify(y) for y in x] if type(x) is list else x);
kw={{key: stringify(value) for key, value in kw.items()}}; {command}(**kw);"""

        args =  casa.strip().split() + casa_opts.strip().split() + ["-c", code, params_string]
        return args





import os
from omegaconf import DictConfig
from scabha.exceptions import ScabhaBaseException


class ConfigurattError(ScabhaBaseException):
    pass


# paths to search for _include statements
PATH = ['.']

# package version info stored with code dependencies
PACKAGE_VERSION = None


# DictConfig doesn't support pop(), so here's a quick replacement
def pop_conf(conf: DictConfig, key: str, default=None):
    value = conf.get(key, default)
    if key in conf:
        del conf[key]
    return value



import os.path
import sys
import glob
import hashlib
import pathlib
import dill as pickle

import uuid
from dataclasses import make_dataclass

from omegaconf.omegaconf import OmegaConf, DictConfig, ListConfig
from omegaconf.errors import OmegaConfBaseException
from typing import Any, List, Dict, Optional, OrderedDict, Union, Callable

from scabha.exceptions import ScabhaBaseException

from yaml.error import YAMLError
from .deps import ConfigDependencies
from .resolvers import resolve_config_refs
from .common import *

# path for cache
CACHEDIR = os.environ.get("CONFIGURATT_CACHE_DIR") or os.path.expanduser(
    "~/.cache/configuratt")


def _compute_hash(filelist, extra_keys):
    filelist = list(filelist) + list(extra_keys)
    return hashlib.md5(" ".join(filelist).encode()).hexdigest()


def set_cache_dir(cachedir: str):
    global CACHEDIR
    CACHEDIR = cachedir


def clear_cache(log=None):
    if os.path.isdir(CACHEDIR):
        files = glob.glob(f"{CACHEDIR}/*")
        log and log.info(f"clearing {len(files)} cached config(s) from cache")
    else:
        files = []
        log and log.info(f"no configs in cache")
    for filename in files:
        try:
            os.unlink(filename)
        except Exception as exc:
            log and log.error(f"failed to remove cached config {filename}: {exc}")
            sys.exit(1)

def load_cache(filelist: List[str], extra_keys=[], verbose=None):
    filehash = _compute_hash(filelist, extra_keys)
    if not os.path.isdir(CACHEDIR):
        if verbose:
            print(f"{CACHEDIR} does not exist")
        return None, None
    filename = os.path.join(CACHEDIR, filehash)
    if not os.path.exists(filename):
        if verbose:
            print(f"hash file {filename} does not exist")
        return None, None
    # check that all configs are older than the cache
    cache_mtime = os.path.getmtime(filename)
    for f in filelist:
        if os.path.getmtime(f) > cache_mtime:
            if verbose:
                print(f"Config {f} is newer than the cache, forcing reload")
            return None, None
    # load cache
    try:
        conf, deps = pickle.load(open(filename, 'rb'))
        if not isinstance(deps, ConfigDependencies):
            raise TypeError(f"cached deps object is of type {type(deps)}, expecting ConfigDependencies")
    except Exception as exc:
        print(f"Error loading cached config from {filename}: {exc}. Removing the cache.")
        os.unlink(filename)
        return None, None
    # check that all dependencies are older than the cache
    if deps.have_deps_changed(cache_mtime, verbose=verbose):
        return None, None
    if verbose:
        print(f"Loaded cached config for {' '.join(filelist)} from {filename}")
    return conf, deps


def save_cache(filelist: List[str], conf, deps: ConfigDependencies, extra_keys=[], verbose=False):
    pathlib.Path(CACHEDIR).mkdir(parents=True, exist_ok=True)
    filelist = list(filelist)   # add self to dependencies
    filehash = _compute_hash(filelist, extra_keys)
    filename = os.path.join(CACHEDIR, filehash)
    # add ourselves to dependencies, so that cache is cleared if implementation changes
    deps.add(__file__, version=PACKAGE_VERSION)
    pickle.dump((conf, deps), open(filename, "wb"), 2)
    if verbose:
        print(f"Caching config for {' '.join(filelist)} as {filename}")


def load(path: str, use_sources: Optional[List[DictConfig]] = [], name: Optional[str]=None,
            location: Optional[str]=None,
            global_deps: Optional[ConfigDependencies] = None,
            includes: bool=True, selfrefs: bool=True, include_path: str=None,
            use_cache: bool = True, verbose: bool = False):
    """Loads config file, using a previously loaded config to resolve _use references.

    Args:
        path (str): path to config file
        use_sources (Optional[List[DictConfig]]): list of existing configs to be used to resolve "_use" references,
                or None to disable
        name (Optional[str]): name of this config file, used for error messages
        location (Optional[str]): location where this config is being loaded (if not at root level)
        includes (bool, optional): If True (default), "_include" references will be processed
        selfrefs (bool, optional): If False, "_use" references will only be looked up in existing config.
            If True (default), they'll also be looked up within the loaded config.
        include_path (str, optional):
            if set, path to each config file will be included in the section as element 'include_path'

    Returns:
        Tuple of (conf, dependencies)
            conf (DictConfig): config object
            dependencies (ConfigDependencies): filenames that were _included
    """
    conf, dependencies = load_cache((path,), verbose=verbose) if use_cache else (None, None)

    if conf is None:
        subconf = OmegaConf.load(path)
        name = name or os.path.basename(path)
        dependencies = ConfigDependencies()
        dependencies.add(path)
        # include ourself into sources, if _use is in effect, and we've enabled selfrefs
        if use_sources is not None and selfrefs:
            use_sources = [subconf] + list(use_sources)
        conf, deps = resolve_config_refs(subconf, pathname=path, location=location, name=name,
                        includes=includes, use_sources=use_sources, use_cache=False, include_path=include_path)
        # update loaded dependencies with locations
        # deps.update_locations(conf, path)
        # update overall dependencies
        dependencies.update(deps)

        if use_cache:
            save_cache((path,), conf, dependencies, verbose=verbose)

    return conf, dependencies

def load_nested(filelist: List[str],
                structured: Optional[DictConfig] = None,
                typeinfo = None,
                use_sources: Optional[List[DictConfig]] = [],
                location: Optional[str] = None,
                nameattr: Union[Callable, str, None] = None,
                config_class: Optional[str] = None,
                include_path: Optional[str] = None,
                use_cache: bool = True,
                verbose: bool = False):
    """Builds nested configuration from a set of YAML files corresponding to sub-sections

    Parameters
    ----------
    conf : OmegaConf object
        root OmegaConf object to merge content into
    filelist : List[str]
        list of subsection config files to load
    schema : Optional[DictConfig]
        schema to be applied to each file, if any
    use_sources : Optional[List[DictConfig]]
        list of existing configs to be used to resolve "_use" references, or None to disable
    location : Optional[str]
        if set, contents of files are being loaded under 'location.subsection_name'. If not set, then 'subsection_name' is being
        loaded at root level. This is used for correctly formatting error messages and such.
    nameattr : Union[Callable, str, None]
        if None, subsection_name will be taken from the basename of the file. If set to a string such as 'name', will set
        subsection_name from that field in the subsection config. If callable, will be called with the subsection config object as a single
        argument, and must return the subsection name
    config_class : Optional[str]
        name of config dataclass to form (when using typeinfo), if None, then generated automatically
    include_path : Optional[str]
        if set, path to each config file will be included in the section as element 'include_path'

    Returns
    -------
        Tuple of (conf, dependencies)
            conf (DictConfig): config object
            dependencies (set): set of filenames that were _included

    Raises
    ------
    NameError
        If subsection name is not resolved
    """
    section_content, dependencies = load_cache(filelist, verbose=verbose) if use_cache else (None, None)

    if section_content is None:
        section_content = {} # OmegaConf.create()
        dependencies = ConfigDependencies()

        for path in filelist:
            # load file
            subconf, deps = load(path, location=location, use_sources=use_sources, include_path=include_path)
            dependencies.update(deps)
            if include_path:
                subconf[include_path] = path

            # figure out section name
            if nameattr is None:
                name = os.path.splitext(os.path.basename(path))[0]
            elif callable(nameattr):
                name = nameattr(subconf)
            elif nameattr in subconf:
                name = subconf.get(nameattr)
            else:
                raise NameError(f"{path} does not contain a '{nameattr}' field")

            # # resolve _use and _include statements
            # try:
            #     subconf = resolve_config_refs(subconf, f"{location}.{name}" if location else name, conf, subconf))
            # except (OmegaConfBaseException, YAMLError) as exc:
            #     raise ConfigurattError(f"config error in {path}: {exc}")

            # apply schema
            if structured is not None:
                try:
                    subconf = OmegaConf.merge(structured, subconf)
                except (OmegaConfBaseException, YAMLError) as exc:
                    raise ConfigurattError(f"schema error in {path}: {exc}")

            section_content[name] = subconf

        if structured is None and typeinfo is not None:
            if config_class is None:
                config_class = "ConfigClass_" + uuid.uuid4().hex
            fields = [(name, typeinfo) for name in section_content.keys()]
            datacls = make_dataclass(config_class, fields)
            # datacls.__module__ == __name__  # for pickling
            structured = OmegaConf.structured(datacls)
            section_content = OmegaConf.merge(structured, section_content)

        if use_cache:
            save_cache(filelist, section_content, dependencies, verbose=verbose)

    return section_content, dependencies


